# Load packages
from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotDotPlot(dataframe,
                categorical_column_name,
                value_column_name,
                group_column_name,
                # Dot formatting arguments
                color_palette="Paired",
                dot_size=1.5,
                dot_alpha=1,
                connect_dots=True,
                connect_line_color="#666666",
                connect_line_alpha=0.4,
                connect_line_width=0.75,
                # Plot formatting arguments
                display_order_list=None,
                figure_size=(10, 6),
                show_legend=True,
                # Text formatting arguments
                title_for_plot=None,
                subtitle_for_plot=None,
                caption_for_plot=None,
                data_source_for_plot=None,
                title_y_indent=1.15,
                subtitle_y_indent=1.1,
                caption_y_indent=-0.15,
                decimal_places_for_data_label=1,
                data_label_fontsize=11):
    """
    Plot a dot plot with optional connecting lines.

    Args:
        dataframe (pd.DataFrame): The input dataframe.
        categorical_column_name (str): The name of the categorical column in the dataframe.
        value_column_name (str): The name of the value column in the dataframe.
        group_column_name (str): The name of the group column in the dataframe.
        color_palette (str, optional): The color palette to use for the plot. Defaults to "Paired".
        dot_size (float, optional): The size of the dots in the plot. Defaults to 1.5.
        dot_alpha (float, optional): The transparency of the dots in the plot. Defaults to 1.
        connect_dots (bool, optional): Whether to connect the dots with lines. Defaults to True.
        connect_line_color (str, optional): The color of the connecting lines. Defaults to "#666666".
        connect_line_alpha (float, optional): The transparency of the connecting lines. Defaults to 0.4.
        connect_line_width (float, optional): The width of the connecting lines. Defaults to 0.75.
        display_order_list (list, optional): The order in which the categories should be displayed. 
            If not provided, the categories will be sorted by the value column in descending order. Defaults to None.
        figure_size (tuple, optional): The size of the plot figure. Defaults to (10, 6).
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        title_y_indent (float, optional): The y-axis indentation for the title. Defaults to 1.15.
        subtitle_y_indent (float, optional): The y-axis indentation for the subtitle. Defaults to 1.1.
        caption_y_indent (float, optional): The y-axis indentation for the caption. Defaults to -0.15.
        decimal_places_for_data_label (int, optional): The number of decimal places to round the data labels. Defaults to 1.
        data_label_fontsize (int, optional): The fontsize of the data labels. Defaults to 11.

    Raises:
        ValueError: If the display_order_list does not contain all of the categories in the dataframe.

    Returns:
        None
    """
    
    # If display_order_list is provided, check that it contains all of the categories in the dataframe
    if display_order_list != None:
        if not set(display_order_list).issubset(set(dataframe[categorical_column_name].unique())):
            raise ValueError("display_order_list must contain all of the categories in the dataframe.")
    else:
        # If display_order_list is not provided, create one from the dataframe
        display_order_list = dataframe.sort_values(value_column_name, ascending=False)[categorical_column_name]
        display_order_list = display_order_list.unique()
    
    # Initialize the matplotlib figure
    f, ax = plt.subplots(figsize=figure_size)
    
    # If connect_dots is True, create horizontal lines connecting the dots
    if connect_dots == True:
        for i in range(len(display_order_list)):
            # Get the x and y coordinates for the dots
            x_coordinates = dataframe[dataframe[categorical_column_name] == display_order_list[i]][value_column_name]
            y_coordinates = dataframe[dataframe[categorical_column_name] == display_order_list[i]][categorical_column_name]
            
            # Get the x and y coordinates for the lines
            x_line_coordinates = [x_coordinates.min(), x_coordinates.max()]
            y_line_coordinates = [y_coordinates.min(), y_coordinates.max()]
            
            # Plot the lines
            plt.plot(
                x_line_coordinates, 
                y_line_coordinates, 
                color=connect_line_color, 
                alpha=connect_line_alpha, 
                linestyle='dashed', 
                linewidth=connect_line_width
            )
    
    # Create pointplot
    sns.pointplot(
        data=dataframe,
        x=value_column_name, 
        y=categorical_column_name, 
        hue=group_column_name,
        order=display_order_list,
        join=False,
        palette=color_palette, 
        markers='o', 
        scale=dot_size,
        ax=ax
    )
    
    # Remove legend if show_legend is False
    if show_legend == False:
        ax.legend_.remove()
    
    # Add space between the title and the plot
    plt.subplots_adjust(top=0.85)
    
    # Wrap y axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(categorical_column_name, 40, break_long_words=False)) 
    ax.set_ylabel(wrapped_variable_name)
    
    # Format and wrap y axis tick labels using textwrap
    y_tick_labels = ax.get_yticklabels()
    wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label.get_text(), 40, break_long_words=False)) for label in y_tick_labels]
    ax.set_yticklabels(
        wrapped_y_tick_labels,
        fontsize=10, 
        # fontname="Arial", 
        color="#262626"
    )
    
    # Move x-axis to the top
    ax.xaxis.tick_top()
    
    # Change x-axis colors to "#666666"
    ax.tick_params(axis='x', colors="#666666")
    ax.spines['top'].set_color("#666666")
    
    # Remove bottom, left, and right spines
    ax.spines['bottom'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    
    # Add data labels
    abs_values = dataframe.sort_values(value_column_name, ascending=False)[value_column_name].round(decimal_places_for_data_label)
    for p, label in zip(ax.patches, abs_values):
        ax.annotate(
            label,
            (p.get_x() + p.get_width() / 2., p.get_y() + p.get_height() / 2.),
            ha='center',
            va='center',
            fontsize=data_label_fontsize,
            color='#262626',
            xytext=(0, 0),
            textcoords='offset points'
        )
        
    # Set the x indent of the plot titles and captions
    # Get longest y tick label
    longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
    if len(longest_y_tick_label) >= 30:
        x_indent = -0.3
    else:
        x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
        
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
from lifelines import KaplanMeierFitter
from lifelines.statistics import logrank_test, multivariate_logrank_test
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import statsmodels.api as sm
import textwrap

# Declare function
def ConductSurvivalAnalysis(dataframe,
                            outcome_column,
                            time_duration_column,
                            group_column=None,
                            return_time_table=True,
                            plot_survival_curve=True,
                            conduct_log_rank_test=True,
                            significance_level=0.05,
                            print_log_rank_test_results=True,
                            line_color="#3269a8",
                            line_alpha=0.8,
                            sns_color_palette="Set2",
                            add_point_in_time_survival_curve=False,
                            point_in_time_survival_color="#3269a8",
                            title_for_plot="Cumulative Survival Curve",
                            subtitle_for_plot='Shows the cumulative survival probability over time',
                            caption_for_plot=None,
                            data_source_for_plot=None,
                            x_indent=-0.127,
                            title_y_indent=1.125,
                            subtitle_y_indent=1.05,
                            caption_y_indent=-0.3,
                            y_axis_label_indent=0.78,
                            x_axis_label_indent=0.925,
                            figure_size=(8, 6)):
    """
    Conducts survival analysis on a given dataset and returns the results.

    Args:
        dataframe (pandas.DataFrame): The dataset to analyze.
        outcome_column (str): The name of the column containing the outcome variable.
        time_duration_column (str): The name of the column containing the time duration variable.
        group_column (str, optional): The name of the column containing the group variable. Defaults to None.
        return_time_table (bool, optional): Whether to return the time table. Defaults to True.
        plot_survival_curve (bool, optional): Whether to plot the survival curve. Defaults to True.
        conduct_log_rank_test (bool, optional): Whether to conduct the log-rank test. Defaults to True.
        significance_level (float, optional): The significance level for the log-rank test. Defaults to 0.05.
        print_log_rank_test_results (bool, optional): Whether to print the results of the log-rank test. Defaults to True.
        line_color (str, optional): The color of the survival curve. Defaults to "#3269a8".
        line_alpha (float, optional): The alpha value of the survival curve. Defaults to 0.8.
        sns_color_palette (str, optional): The color palette to use for the survival curve. Defaults to "Set2".
        add_point_in_time_survival_curve (bool, optional): Whether to add a point-in-time survival curve. Defaults to False.
        point_in_time_survival_color (str, optional): The color of the point-in-time survival curve. Defaults to "#3269a8".
        title_for_plot (str, optional): The title of the survival curve plot. Defaults to "Cumulative Survival Curve".
        subtitle_for_plot (str, optional): The subtitle of the survival curve plot. Defaults to "Shows the cumulative survival probability over time".
        caption_for_plot (str, optional): The caption of the survival curve plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the survival curve plot. Defaults to None.
        x_indent (float, optional): The x-axis indent of the survival curve plot. Defaults to -0.127.
        title_y_indent (float, optional): The title y-axis indent of the survival curve plot. Defaults to 1.125.
        subtitle_y_indent (float, optional): The subtitle y-axis indent of the survival curve plot. Defaults to 1.05.
        caption_y_indent (float, optional): The caption y-axis indent of the survival curve plot. Defaults to -0.3.
        y_axis_label_indent (float, optional): The y-axis label indent of the survival curve plot. Defaults to 0.78.
        x_axis_label_indent (float, optional): The x-axis label indent of the survival curve plot. Defaults to 0.925.
        figure_size (tuple, optional): The size of the survival curve plot. Defaults to (8, 6).

    Returns:
        pandas.DataFrame: The time table, if return_time_table is True.
    """
    
    # Select the columns to keep
    if group_column is None:
        dataframe = dataframe[[outcome_column, time_duration_column]].copy()
    else:
        dataframe = dataframe[[outcome_column, time_duration_column, group_column]].copy()

    # Remove records with missing values
    dataframe = dataframe.dropna()

    # Order the dataframe by outcome
    dataframe = dataframe.sort_values(by = [outcome_column, time_duration_column], ascending = [True, True])
    
    # Create a new column showing unqiue status as group starting with 0
    unique_status = dataframe[outcome_column].unique()
    dataframe["Survival Group"] = dataframe[outcome_column].apply(lambda x: list(unique_status).index(x))
    # Print the survival group and their corresponding status
    print("Outcome variables have been assigned accordingly to each group:")
    print(dataframe[["Survival Group", outcome_column]].drop_duplicates().reset_index(drop = True))

    # Calculate the survival probability (Kaplan-Meier Estimator)
    if group_column is None:
        model = KaplanMeierFitter()
        model.fit(
            durations=dataframe[time_duration_column], 
            event_observed=dataframe[outcome_column]
        )
    else:
        # Create dictionary to hold models for each group
        dict_models = {}
        for group in dataframe[group_column].unique():
            # Separate the dataframe by group
            group_data = dataframe[dataframe[group_column] == group].copy()
            # Fit the model for each group
            dict_models[group] = KaplanMeierFitter()
            dict_models[group].fit(
                durations=group_data[time_duration_column], 
                event_observed=group_data[outcome_column]
            )

    # Create a new dataframe to store the time table/survival probability
    if group_column is None:
        # Get event table from model
        data_time_table = model.event_table
        data_time_table = data_time_table.reset_index(drop=False)
        data_time_table.columns = data_time_table.columns.str.title()
        data_time_table.columns = data_time_table.columns.str.replace("_", " ")
        data_time_table = data_time_table.rename(columns={"Event At": time_duration_column})
        # Calculate the survival probability
        data_time_table["Survival Probability - Point in Time"] = (data_time_table["At Risk"] - data_time_table["Observed"]) / data_time_table["At Risk"]
        # Calculate the cumulative survival probability
        data_time_table["Survival Probability - Cumulative"] = model.survival_function_.values
        # Add confidence interval 
        data_time_table[["Survival Probability - Cumulative (Lower 95% C.I.)",  "Survival Probability - Cumulative (Upper 95% C.I.)"]] = model.confidence_interval_survival_function_.values
    else:
        # Sort the dataframe by group and time duration
        dataframe = dataframe.sort_values(by = [group_column, time_duration_column], ascending = [True, True])
        # Create dataframe to store the time table/survival probability
        data_time_table = pd.DataFrame()
        for group in dataframe[group_column].unique():
            group_model = dict_models[group]
            # Get event table from model
            group_data_time_table = group_model.event_table
            group_data_time_table = group_data_time_table.reset_index(drop=False)
            group_data_time_table.columns = group_data_time_table.columns.str.title()
            group_data_time_table.columns = group_data_time_table.columns.str.replace("_", " ")
            group_data_time_table = group_data_time_table.rename(columns={"Event At": time_duration_column})
            # Calculate the survival probability
            group_data_time_table["Survival Probability - Point in Time"] = (group_data_time_table["At Risk"] - group_data_time_table["Observed"]) / group_data_time_table["At Risk"]
            # Calculate the cumulative survival probability
            group_data_time_table["Survival Probability - Cumulative"] = group_model.survival_function_.values
            # Add confidence interval 
            group_data_time_table[["Survival Probability - Cumulative (Lower 95% C.I.)",  "Survival Probability - Cumulative (Upper 95% C.I.)"]] = group_model.confidence_interval_survival_function_.values
            # Add group column
            group_data_time_table[group_column] = group
            # Row bind to the dataframe
            data_time_table = pd.concat([data_time_table, group_data_time_table], axis=0).reset_index(drop=True)
            
    # If plot_survival_curve is True, plot the survival curve
    if plot_survival_curve:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Use Seaborn to create a line plot
        if group_column is None:
            sns.lineplot(
                data=data_time_table,
                x=time_duration_column,
                y="Survival Probability - Cumulative",
                color=line_color,
                alpha=line_alpha,
                ax=ax
            )
            # Add point in time curve, if requested
            if add_point_in_time_survival_curve:
                sns.lineplot(
                    data=data_time_table,
                    x=time_duration_column,
                    y="Survival Probability - Point in Time",
                    color=line_color,
                    alpha=line_alpha,
                    ax=ax
                )
            # Add confidence interval as a shaded area
            plt.fill_between(
                data_time_table[time_duration_column].values, 
                data_time_table["Survival Probability - Cumulative (Upper 95% C.I.)"].values,
                data_time_table["Survival Probability - Cumulative (Lower 95% C.I.)"].values,
                alpha=0.2,
                color=line_color
            )
        else:
            sns.lineplot(
                data=data_time_table,
                x=time_duration_column,
                y="Survival Probability - Cumulative",
                hue=group_column,
                palette=sns_color_palette,
                alpha=line_alpha,
                ax=ax
            )
            # Add confidence interval as a shaded area
            for i in range(data_time_table[group_column].unique().shape[0]):
                # Get current group
                group = data_time_table[group_column].unique().tolist()[i]
                # Get group color
                group_color = sns.color_palette(sns_color_palette)[i]
                # Filter to current group
                group_data_time_table = data_time_table[data_time_table[group_column] == group]
                # Add shaded area
                plt.fill_between(
                    group_data_time_table[time_duration_column].values, 
                    group_data_time_table["Survival Probability - Cumulative (Upper 95% C.I.)"].values,
                    group_data_time_table["Survival Probability - Cumulative (Lower 95% C.I.)"].values,
                    alpha=0.2,
                    color=group_color
                )
        
        # Remove top and right spines, and set bottom and left spines to gray
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')

        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
        ax.yaxis.set_label_coords(-0.1, y_axis_label_indent)
        ax.yaxis.set_label_text(
            "Survival Probability - Cumulative",
            fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
        ax.xaxis.set_label_coords(x_axis_label_indent, -0.1)
        ax.xaxis.set_label_text(
            time_duration_column,
            fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                if len(caption_for_plot) > 120:
                    # Split the caption into words
                    words = caption_for_plot.split(" ")
                    # Initialize the wrapped caption
                    wrapped_caption = ""
                    # Initialize the line length
                    line_length = 0
                    # Iterate through the words
                    for word in words:
                        # If the word is too long to fit on the current line, add a new line
                        if line_length + len(word) > 120:
                            wrapped_caption = wrapped_caption + "\n"
                            line_length = 0
                        # Add the word to the line
                        wrapped_caption = wrapped_caption + word + " "
                        # Update the line length
                        line_length = line_length + len(word) + 1
            else:
                wrapped_caption = ""
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )

        # Show plot
        plt.show()

    # If there is a group column, run a log-rank test for each pair of groups
    if conduct_log_rank_test:
        if group_column != None:
            groups = dataframe[group_column].unique()
            # If there are more than two groups, run a multivariate log-rank test
            if len(groups) >= 3:
                results = multivariate_logrank_test(
                    dataframe[time_duration_column],
                    dataframe[group_column],
                    dataframe[outcome_column]
                )
                if print_log_rank_test_results:
                    results.print_summary()
            # If there are two groups, run a bivariate log-rank test
            elif len(groups) == 2:
                results = logrank_test(
                    dataframe[dataframe[group_column] == groups[0]][time_duration_column],
                    dataframe[dataframe[group_column] == groups[1]][time_duration_column],
                    dataframe[dataframe[group_column] == groups[0]][outcome_column],
                    dataframe[dataframe[group_column] == groups[1]][outcome_column]
                )
                if print_log_rank_test_results:
                    results.print_summary()
    
    # Create a dictionary to return models and survival tables
    return_dict = {}
    if group_column != None:
        for key, value in dict_models.items():
            return_dict[key] = value
    return_dict["survival_table"] = data_time_table
    
    # Return the dictionary
    return return_dict




# Load packages
from sklearn.preprocessing import KBinsDiscretizer
import numpy as np
import pandas as pd

# Declare function
def CreateBinnedColumn(dataframe,
                       numeric_column_name,
                       number_of_bins=6,
                       binning_strategy='kmeans',
                       new_column_name=None):
    """
    This function creates a binned column in a dataframe based on a numeric variable. 
    The default is to create 6 bins using the k-means binning strategy. 
    The new column name is the name of the variable to bin with '- Binned' appended to it.

    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        numeric_column_name (str): The name of the variable to bin. Should be a numeric variable.
        number_of_bins (int, optional): The number of bins to create. Defaults to 6.
        binning_strategy (str, optional): The strategy to use to create the bins. Defaults to 'kmeans'.
        new_column_name (str, optional): The name of the new column containing the bins. Defaults to None.

    Returns:
        Pandas dataframe: An updated Pandas dataframe with the bins for the variable to bin
    """
    
    # Create new column name if not provided
    if new_column_name is None:
        new_column_name = numeric_column_name + '- Binned'
        
    # Select only the variable to bin
    dataframe_bin = dataframe[[numeric_column_name]].copy()
    
    # Keep complete cases only
    dataframe_bin = dataframe_bin.dropna()
    
    # Keep only finite values
    dataframe_bin = dataframe_bin[np.isfinite(dataframe_bin).all(1)]
        
    # Create the array to bin
    array_to_bin = np.array(dataframe_bin[numeric_column_name]).reshape(-1, 1)
    
    # Create the discretizer
    binner = KBinsDiscretizer(n_bins=number_of_bins, 
                              strategy=binning_strategy,
                              encode='ordinal')
    
    # Fit the discretizer
    binner.fit(array_to_bin)
    
    # Create the binned variable
    dataframe_bin[new_column_name] = binner.transform(array_to_bin)
    
    # Merge the binned variable back into the original dataframe
    dataframe = dataframe.merge(
        dataframe_bin[[new_column_name]],
        how='left',
        right_index=True,
        left_index=True
    )
    
    # Return the dataframe
    return dataframe




# Load packages
import matplotlib.pyplot as plt
import numpy as np

# Declare function
def VisualizeMatrixAsLinearTransformation(two_by_two_matrix1,
                                          two_by_two_matrix2=None,
                                          plot_with_grid=False,
                                          show_labels=True,
                                          matrix1_color='#033dfc',
                                          matrix2_color='#03b1fc',
                                          x_min=-5,
                                          x_max=5,
                                          y_min=-5,
                                          y_max=5):
    """
    Plots up to two 2x2 matrices and the unit square and the transformed unit square. 
    If a second matrix is provided, it will plot the transformed unit square twice, once for each matrix.
    
    Args:
        two_by_two_matrix1 (list or np.array): A 2x2 matrix. If a list is provided, it will be converted to a numpy array.
        two_by_two_matrix2 (list or np.array, optional): A 2x2 matrix. If a list is provided, it will be converted to a numpy array. Defaults to None.
        plot_with_grid (bool, optional): Whether or not to plot a grid on plot. Defaults to False.
        show_labels (bool, optional): Whether or not to show labels on plot. Defaults to True.
        matrix1_color (str, optional): Color of the first matrix. Defaults to '#033dfc'.
        matrix2_color (str, optional): Color of the second matrix. Defaults to '#03b1fc'.
        x_min (int, optional): The minimum x value to plot. Defaults to -5.
        x_max (int, optional): The maximum x value to plot. Defaults to 5.
        y_min (int, optional): The minimum y value to plot. Defaults to -5.
        y_max (int, optional): The maximum y value to plot. Defaults to 5.
    """
    
    # If matrix is list, convert to numpy array
    if type(two_by_two_matrix1) == list:
        two_by_two_matrix1 = np.array(two_by_two_matrix1)
    if two_by_two_matrix2 is not None:
        if type(two_by_two_matrix2) == list:
            two_by_two_matrix2 = np.array(two_by_two_matrix2)
            
    # Perform matrix multiplication
    transformed_square = np.dot(two_by_two_matrix1, np.array([[0,1,1,0,0],[0,0,1,1,0]]))
    if two_by_two_matrix2 is not None:
        transformed_square_2 = np.dot(two_by_two_matrix2, transformed_square)
        
    # Ensure matrix is 2x2
    if two_by_two_matrix1.shape != (2,2):
        raise Exception("Matrix #1 must be 2x2")
    if two_by_two_matrix2 is not None:
        if two_by_two_matrix1.shape != (2,2):
            raise Exception("Matrix #2 must be 2x2")
    
    # Plot the vectors
    _, ax = plt.subplots(figsize=(10, 10))
    ax.tick_params(axis='x', labelsize=14)
    ax.tick_params(axis='y', labelsize=14)
    
    # Set the x and y limits
    ax.set_xlim([x_min-1, x_max+1])
    ax.set_ylim([y_min-1, y_max+1])

    # Add a grid, if requested
    if plot_with_grid:
        plt.grid()
    
    # Set aspect ratio to equal
    plt.gca().set_aspect("equal")
    
    # Remove borders
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.gca().spines['bottom'].set_visible(False)
    plt.gca().spines['left'].set_visible(False)
    
    # Add line at x=0 and y=0
    plt.axhline(y=0, color='k', alpha=0.25)
    plt.axvline(x=0, color='k', alpha=0.25)
    
    # Plot the unit square
    plt.plot([0,1,1,0,0],[0,0,1,1,0],
             color='black',
             alpha=0.5)
    plt.fill([0,1,1,0,0],[0,0,1,1,0],
             color='black',
             alpha=0.1)
    # Add label for unit square
    if show_labels:
        plt.text(
            0.5, 
            -0.15, 
            "Unit Square", 
            fontsize=8, 
            fontfamily='Arial', 
            ha='center', 
            va='center'
        )
    
    # Plot the transformed unit square
    plt.plot(
        transformed_square[0,:],
        transformed_square[1,:],
        color=matrix1_color,
        alpha=0.25
    )
    plt.fill(
        transformed_square[0,:],
        transformed_square[1,:],
        color=matrix1_color,
        alpha=0.1
    )
    # Add label for transformed unit square
    if show_labels:
        plt.text(
            transformed_square[0,2]+0.15,
            transformed_square[1,2]+0.15, 
            "Transformed Unit Square #1", 
            fontsize=8, 
            fontfamily='Arial', 
            ha='center', 
            va='center', 
            color=matrix1_color
        )
    
    # If a second matrix is provided, plot the transformed unit square
    if two_by_two_matrix2 is not None:
        # Plot the transformed unit square
        plt.plot(
            transformed_square_2[0,:],
            transformed_square_2[1,:],
            color=matrix2_color,
            alpha=0.25
        )
        plt.fill(
            transformed_square_2[0,:],
            transformed_square_2[1,:],
            color=matrix2_color,
            alpha=0.1
        )
        # Add label for transformed unit square
        if show_labels:
            plt.text(
                transformed_square_2[0,2]+0.15,
                transformed_square_2[1,2]+0.15, 
                "Transformed Unit Square #2", 
                fontsize=8, 
                fontfamily='Arial', 
                ha='center', 
                va='center', 
                color=matrix2_color
            )
    
    # Add title
    plt.title("Linear Transformation", fontsize=20, fontfamily='Arial')
    
    # Show the plot
    plt.show()




# Load packages
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

# Declare function
def SolveSystemOfEquations(coefficients,
                           constants=None,
                           show_plot=True,
                           plot_boundary=10):
    """
    Plots a system of equations.
    
    Args:
        coefficients (list of lists): A matrix of coefficients for the system of equations
        constants (list): A 1D array of constants for the system of equations
    """ 
    
    # Convert coefficients to a numpy array
    coefficients = np.array(coefficients)
    
    # If constants are not provided, set them to 0
    if constants is None:
        constants = np.zeros(coefficients.shape[0])
    else:
        constants = np.array(constants)
    
    # Calculate the determinant of the coefficient matrix
    determinant = np.linalg.det(coefficients)
    print("Determinant:", '{:f}'.format(determinant))
    if determinant == 0:
        print("The system of equations is singular, and does not have a unique solution. At least two equations are linearly dependent.")
    else:
        print("The system of equations is non-singular, and has a unique solution. The equations are linearly independent.")
    
    # Solve the system of equations
    if determinant != 0:
        solution = np.linalg.solve(coefficients, constants)
        print("Solution:", str(solution))
    
    if show_plot: 
        if coefficients.shape[1] == 2:
            # Plot the system of equations
            for i in range(len(constants)):
                x = np.linspace(-1*plot_boundary, plot_boundary, 100)
                y = (constants[i] - coefficients[i, 0] * x) / coefficients[i, 1]
                
                # Generate name for the equation
                equation_name = str(coefficients[i, 0]) + "x + " + str(coefficients[i, 1]) + "y = " + str(constants[i])
                plt.plot(x, y, label=equation_name)
        
            # Create legend
            plt.legend()
            
            # Show plot
            plt.show()
        
        else:
            print("\nCannot plot system of equations with more than 2 variables.")




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotCorrelationMatrix(dataframe,
                          list_of_value_column_names,
                          list_of_outcome_column_names=None,
                          show_as_pairplot=True,
                          # Pairplot formatting options
                          pairplot_size=(20, 20),
                          scatter_fill_color="#3269a8",
                          # Line formatting options
                          fit_lowess_line=True,
                          line_color="#cc4b5a",
                          # Text formatting arguments
                          title_for_plot="Pairplot of Numeric Variables",
                          subtitle_for_plot="Shows the linear relationship between numeric variables",
                          caption_for_plot=None,
                          data_source_for_plot=None,
                          x_indent=-0.7,
                          title_y_indent=1.12,
                          title_font_size=14,
                          subtitle_y_indent=1.03,
                          subtitle_font_size=11,
                          caption_y_indent=-0.35,
                          caption_font_size=8):
    """
    This function plots a correlation matrix or pairplot for a given dataframe. 

    Args:
        dataframe (pandas.DataFrame): The dataframe to plot.
        list_of_value_column_names (list): The names of the columns to include in the plot.
        list_of_outcome_column_names (list, optional): The names of the outcome columns to include in the plot. Defaults to None.
        show_as_pairplot (bool, optional): Whether to show the plot as a pairplot. Defaults to True.
        pairplot_size (tuple, optional): The size of the pairplot. Defaults to (20, 20).
        scatter_fill_color (str, optional): The fill color for scatter plots. Defaults to "#3269a8".
        fit_lowess_line (bool, optional): Whether to fit a lowess line to the data. Defaults to True.
        line_color (str, optional): The color of the line. Defaults to "#cc4b5a".
        title_for_plot (str, optional): The title for the plot. Defaults to "Pairplot of Numeric Variables".
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to "Shows the linear relationship between numeric variables".
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        x_indent (float, optional): The x-indent for the plot. Defaults to -0.7.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.12.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.03.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.35.
    """
    
    # Select relevant variables, keep complete cases only
    if list_of_outcome_column_names is None:
        completed_df = dataframe[list_of_value_column_names].dropna()
    else:
        completed_df = dataframe[list_of_outcome_column_names + list_of_value_column_names].dropna()

    # Drop Inf values
    completed_df = completed_df[np.isfinite(completed_df).all(1)]
    print("Count of complete observations for correlation matrix: " + str(len(completed_df.index)))
    
    # Show pairplot if specified -- otherwise, print correlation matrix
    if show_as_pairplot:
        plt.figure(figsize=pairplot_size)
        if list_of_outcome_column_names is None:
            ax = sns.pairplot(
                data=completed_df,
                kind='reg',
                markers='o',
                diag_kws={
                    'color': scatter_fill_color,
                    'alpha': 0.80
                },
                plot_kws={
                    'lowess': fit_lowess_line,
                    'line_kws': {
                        'color': line_color,
                        'alpha': 0.80
                    },
                    'scatter_kws': {
                        'color': scatter_fill_color,
                        'alpha': 0.30
                    }
                }
            )
            
        else:
            ax = sns.pairplot(
                data=completed_df,
                x_vars=list_of_value_column_names,
                y_vars=list_of_outcome_column_names,
                kind='reg',
                markers='o',
                diag_kws={
                    'color': scatter_fill_color,
                    'alpha': 0.80
                },
                plot_kws={
                    'lowess': fit_lowess_line,
                    'line_kws': {
                        'color': line_color,
                        'alpha': 0.80
                    },
                    'scatter_kws': {
                        'color': scatter_fill_color,
                        'alpha': 0.30
                    }
                }
            )
            
        # Wrap labels in each subplot
        for ax in ax.axes.flat:
            # x-axis labels
            x_labels = ax.get_xlabel()
            wrapped_x_labels = "\n".join(textwrap.wrap(x_labels, 40, break_long_words=False))
            ax.set_xlabel(wrapped_x_labels, horizontalalignment='center')
            # y-axis labels
            y_labels = ax.get_ylabel()
            wrapped_y_labels = "\n".join(textwrap.wrap(y_labels, 40, break_long_words=False))
            ax.set_ylabel(wrapped_y_labels, horizontalalignment='center')
        
        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Adjust y-indents for title and subtitle
        if list_of_outcome_column_names is None:
            title_y_indent = title_y_indent*len(list_of_value_column_names)
            subtitle_y_indent = subtitle_y_indent*len(list_of_value_column_names)
        else:
            title_y_indent = title_y_indent*len(list_of_outcome_column_names)
            subtitle_y_indent = subtitle_y_indent*len(list_of_outcome_column_names)
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent*len(list_of_value_column_names),
            y=title_y_indent,
            s=title_for_plot,
            # fontname="Arial",
            fontsize=title_font_size,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent*len(list_of_value_column_names),
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            # fontname="Arial",
            fontsize=subtitle_font_size,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                if list_of_outcome_column_names is None:
                    wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                else:
                    wrapped_caption = textwrap.fill(caption_for_plot, 60, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent*len(list_of_value_column_names),
                y=caption_y_indent,
                s=wrapped_caption,
                # fontname="Arial",
                fontsize=caption_font_size,
                color="#666666",
                transform=ax.transAxes
            )
    
    else:
        print(completed_df.corr())




# Load packages
import pandas as pd
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def AddTukeyOutlierColumn(dataframe, 
                          value_column_name,
                          tukey_boundary_multiplier=1.5,
                          plot_tukey_outliers=False,
                          # Histogram formatting arguments
                          fill_color="#999999",
                          fill_transparency=0.6,
                          # Text formatting arguments
                          title_for_plot="Tukey Outliers",
                          subtitle_for_plot="Shows the values that are Tukey outliers.",
                          caption_for_plot=None,
                          data_source_for_plot=None,
                          show_y_axis=False,
                          title_y_indent=1.1,
                          subtitle_y_indent=1.05,
                          caption_y_indent=-0.15,
                          # Plot formatting arguments
                          figure_size=(8, 6)):
    """
    Adds a column to the DataFrame that flags Tukey outliers.

    Args:
        dataframe (pandas.DataFrame): The DataFrame containing the data.
        value_column_name (str): The name of the column to check for outliers.
        tukey_boundary_multiplier (float, optional): The number of IQRs to use for the Tukey outlier boundary. Defaults to 1.5.
        plot_tukey_outliers (bool, optional): Whether to plot the Tukey outliers. Defaults to False.
        fill_color (str, optional): The color to fill the histogram bars with. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the histogram bars. Defaults to 0.6.
        title_for_plot (str, optional): The title of the plot. Defaults to "Tukey Outliers".
        subtitle_for_plot (str, optional): The subtitle of the plot. Defaults to "Shows the values that are Tukey outliers.".
        caption_for_plot (str, optional): The caption of the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis. Defaults to False.
        title_y_indent (float, optional): The y-indent of the title. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent of the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent of the caption. Defaults to -0.15.
        figure_size (tuple, optional): The size of the plot figure. Defaults to (8, 6).

    Returns:
        pandas.DataFrame: The original DataFrame with an added column for outlier flags.
    """
    
    # Calculate the IQR
    Q1 = dataframe[value_column_name].quantile(0.25)
    Q3 = dataframe[value_column_name].quantile(0.75)
    IQR = Q3 - Q1

    # Define the outlier bounds
    lower_bound = Q1 - tukey_boundary_multiplier * IQR
    upper_bound = Q3 + tukey_boundary_multiplier * IQR

    # Flag outliers
    dataframe[f'{value_column_name} - Tukey outlier'] = ((dataframe[value_column_name] < lower_bound) | (dataframe[value_column_name] > upper_bound)).astype(int)
    
    # Plot outliers, if requested
    if plot_tukey_outliers:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        sns.histplot(
            data=dataframe,
            x=value_column_name,
            color=fill_color,
            alpha=fill_transparency,
            hue=f'{value_column_name} - Tukey outlier',
            palette={True: "#FF0000", False: "#999999"}
        )
        
        # Remove the legend
        ax.get_legend().remove()
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Show the lower boundary as a vertical line with a label
        ax.axvline(
            x=lower_bound,
            ymax=0.97-.02,
            color="#262626",
            linestyle="--",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=lower_bound, 
            y=plt.ylim()[1] * 0.97, 
            s='Lower bound: {:.2f}'.format(lower_bound),
            horizontalalignment='center',
            fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
        
        # Show the upper boundary as a vertical line with a label
        ax.axvline(
            x=upper_bound,
            ymax=0.97-.02,
            color="#262626",
            linestyle="--",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=upper_bound, 
            y=plt.ylim()[1] * 0.97, 
            s='Upper bound: {:.2f}'.format(upper_bound),
            horizontalalignment='center',
            fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
    
    # Return the updated dataframe
    return dataframe




# Load packages
import PyPDF2
import re

# Delclare function
def ExtractTextFromPDF(filepath_to_pdf,
                       filepath_for_exported_text,
                       start_page=1,
                       end_page=None,
                       show_word_count=False,
                       show_estimated_token_count=False):
    """
    This function extracts text from a PDF file, cleans it, then saves it to a text file.

    Args:
        filepath_to_pdf (str): The path to the PDF file that you want to extract text from.
        filepath_for_exported_text (str): The path to the text file that you want to save the extracted text to.
        start_page (int, optional): The page number of the PDF file that you want to start extracting text from. Defaults to 1.
        end_page (int or None, optional): The page number of the PDF file that you want to stop extracting text from. If None, the function will extract text from all pages in the PDF file. Defaults to None.
        show_word_count (bool, optional): Whether or not to show the word count of the extracted text. Defaults to True.
    """
    
    # Ensure that filepath_to_pdf is a string ending in .pdf
    if not isinstance(filepath_to_pdf, str) or not filepath_to_pdf.endswith('.pdf'):
        raise TypeError("filepath_to_pdf must be a string ending in .pdf")  
    
    # Ensure that filepath_for_exported_text is a string ending in .txt
    if not isinstance(filepath_for_exported_text, str) or not filepath_for_exported_text.endswith('.txt'):
        raise TypeError("filepath_for_exported_text must be a string ending in .txt")
    
    # Open the PDF file in read-binary mode
    with open(filepath_to_pdf, 'rb') as pdf_file:
        # Create a PDF reader object
        pdf_reader = PyPDF2.PdfReader(pdf_file)
        
        # If end_page is not specified, set it to the number of pages in the PDF
        if end_page == None:
            # Get the number of pages in the PDF document
            num_pages = len(pdf_reader.pages)
            end_page = num_pages

        # Create a new text file to write the extracted text to
        with open(filepath_for_exported_text, 'w', encoding='utf-8') as text_file:
            # Loop through each page in the PDF document
            for page_num in range(start_page-1, end_page):
                # Get the page object for the current page
                page_obj = pdf_reader.pages[page_num]

                # Extract the text from the current page
                page_text = page_obj.extract_text()
                
                # Clean up the text
                page_text = page_text.strip()
                page_text = page_text.replace("  ", " ")
                page_text = page_text.replace(" -", "-")
                
                # Use regex to replace all instances of a space and new line between to letters
                page_text = re.sub(r'(?<=[a-zA-Z]) \n(?=[a-zA-Z])', '', page_text)
                
                # Write the text from the current page to the text file
                text_file.write(page_text)

    # Print a message to indicate that the text has been extracted and saved
    print("Text extracted and saved to " + filepath_for_exported_text + ".")
    
    # Read the text file if word count or estimated token count is True
    if show_word_count or show_estimated_token_count:
        final_text = open(filepath_for_exported_text, 'r', encoding='utf-8').read()
        
    # Show the word count if show_word_count is True
    if show_word_count:
        # Split the text into a list of words
        final_text = final_text.split()
        # Get the number of words in the text
        word_count = len(final_text)
        # Print the number of words in the text
        print(f"Word count: {word_count}")
        
    # Show the estimated token count if show_estimated_token_count is True
    if show_estimated_token_count:
        # Get the number of words in the text
        word_count = len(final_text)
        # Get the number of tokens in the text
        number_of_tokens = word_count * (100/75)
        print(f"Estimated token count: {int(round(number_of_tokens, 0))}")



# Load packages
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def PlotTimeSeries(dataframe,
                   value_column_name,
                   time_column_name,
                   grouping_column_name=None,
                   # Line formatting arguments
                   line_color="#3269a8",
                   line_alpha=0.8,
                   color_palette="Set2",
                   show_legend=True,
                   # Marker formatting arguments
                   marker_size_column_name=None,
                   marker_size_normalization=None,
                   # Text formatting arguments
                   number_of_x_axis_ticks=None,
                   x_axis_tick_rotation=None,
                   title_for_plot=None,
                   subtitle_for_plot=None,
                   caption_for_plot=None,
                   data_source_for_plot=None,
                   x_indent=-0.127,
                   title_y_indent=1.125,
                   subtitle_y_indent=1.05,
                   caption_y_indent=-0.3,
                   # Plot formatting arguments
                   figure_size=(8, 5)):
    """
    This function plots a time series from a given dataframe.

    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be plotted.
        value_column_name (str): The name of the column in the dataframe that contains the values to be plotted.
        time_column_name (str): The name of the column in the dataframe that contains the time data.
        grouping_column_name (str, optional): The name of the column in the dataframe to group data by. Defaults to None.
        line_color (str, optional): The color of the line in the plot. Defaults to "#3269a8".
        line_alpha (float, optional): The transparency of the line in the plot. Defaults to 0.8.
        color_palette (str, optional): The color palette to use for the plot. Defaults to "Set2".
        marker_size_column_name (str, optional): The name of the column in the dataframe that contains the marker size data. Defaults to None.
        marker_size_normalization (tuple, optional): The normalization factor for the marker size data. Defaults to None.
        number_of_x_axis_ticks (int, optional): The number of ticks on the x-axis. Defaults to None.
        x_axis_tick_rotation (int, optional): The rotation of the x-axis ticks. Defaults to None.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        x_indent (float, optional): The x-indent for the plot text. Defaults to -0.127.
        title_y_indent (float, optional): The y-indent for the plot title. Defaults to 1.125.
        subtitle_y_indent (float, optional): The y-indent for the plot subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the plot caption. Defaults to -0.3.
        figure_size (tuple, optional): The size of the figure for the plot. Defaults to (8, 5).
    """
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Use Seaborn to create a line plot
    if grouping_column_name is None:
        sns.lineplot(
            data=dataframe,
            x=time_column_name,
            y=value_column_name,
            color=line_color,
            alpha=line_alpha,
            ax=ax
        )
    else:
        sns.lineplot(
            data=dataframe,
            x=time_column_name,
            y=value_column_name,
            hue=grouping_column_name,
            palette=color_palette,
            alpha=line_alpha,
            ax=ax
        )
        
    # Add scatterplot markers if a column name is provided
    if marker_size_column_name is not None:
        if grouping_column_name is None:
            sns.scatterplot(
                data=dataframe,
                x=time_column_name,
                y=value_column_name,
                color=line_color,
                size=marker_size_column_name,
                size_norm=marker_size_normalization,
                alpha=line_alpha,
                ax=ax,
                linewidth=0.5,
                edgecolor="white",
                legend=False
            )
        else:
            sns.scatterplot(
                data=dataframe,
                x=time_column_name,
                y=value_column_name,
                hue=grouping_column_name,
                size=marker_size_column_name,
                size_norm=marker_size_normalization,
                palette=color_palette,
                alpha=line_alpha,
                ax=ax,
                linewidth=0.5,
                edgecolor="white",
                legend=False
            )
    
    # Remove top and right spines, and set bottom and left spines to gray
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_color('#666666')
    ax.spines['left'].set_color('#666666')
    
    # If show_legend is False, remove the legend
    if show_legend == False:
        ax.legend_.remove()
    
    # Set the number of ticks on the x-axis
    if number_of_x_axis_ticks is not None:
        ax.xaxis.set_major_locator(plt.MaxNLocator(number_of_x_axis_ticks))
        
    # Rotate x-axis tick labels
    if x_axis_tick_rotation is not None:
        plt.xticks(rotation=x_axis_tick_rotation)

    # Format tick labels to be Arial, size 9, and color #666666
    ax.tick_params(
        which='major',
        labelsize=9,
        color='#666666'
    )
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Word wrap the subtitle without splitting words
    if subtitle_for_plot != None:   
        subtitle_for_plot = textwrap.fill(subtitle_for_plot, 100, break_long_words=False)
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            # fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
    
    # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
    ax.yaxis.set_label_coords(-0.1, 0.84)
    ax.yaxis.set_label_text(
        value_column_name,
        # fontname="Arial",
        fontsize=10,
        color="#666666"
    )
    
    # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
    ax.xaxis.set_label_coords(0.9, -0.1)
    ax.xaxis.set_label_text(
        time_column_name,
        # fontname="Arial",
        fontsize=10,
        color="#666666"
    )
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )

    # Show plot
    plt.show()
    



from .ExtractTextFromPDF import ExtractTextFromPDF
from .FetchPDFFromURL import FetchPDFFromURL
from .FetchUSShapefile import FetchUSShapefile
from .FetchWebsiteText import FetchWebsiteText
from .GetGoogleSearchResults import GetGoogleSearchResults
from .GetZipFile import GetZipFile



# Load packages
import pandas as pd
import os
from math import ceil
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotSingleVariableHistogram(dataframe,
                                value_column_name,
                                # Histogram formatting arguments
                                fill_color="#999999",
                                fill_transparency=0.6,
                                show_mean=True,
                                show_median=True,
                                # Text formatting arguments
                                title_for_plot=None,
                                subtitle_for_plot=None,
                                caption_for_plot=None,
                                data_source_for_plot=None,
                                show_y_axis=False,
                                title_y_indent=1.1,
                                subtitle_y_indent=1.05,
                                caption_y_indent=-0.15,
                                # Plot formatting arguments
                                figure_size=(8, 6)):
    """
    This function plots a single variable histogram from a given dataframe.

    Args:
        dataframe (pandas.DataFrame): The dataframe containing the data to be plotted.
        value_column_name (str): The name of the column in the dataframe to be plotted.
        fill_color (str, optional): The fill color for the histogram. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the fill color. Defaults to 0.6.
        show_mean (bool, optional): Whether to show the mean on the plot. Defaults to True.
        show_median (bool, optional): Whether to show the median on the plot. Defaults to True.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis. Defaults to False.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
        figure_size (tuple, optional): The size of the figure. Defaults to (8, 6).
    """
    
    # Check that the column exists in the dataframe.
    if value_column_name not in dataframe.columns:
        raise ValueError("Column {} does not exist in dataframe.".format(value_column_name))

    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Create histogram using seaborn
    sns.histplot(
        data=dataframe,
        x=value_column_name,
        color=fill_color,
        alpha=fill_transparency,
    )
    
    # Remove top, left, and right spines. Set bottom spine to dark gray.
    ax.spines['top'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_color("#262626")
    
    # Remove the y-axis, and adjust the indent of the plot titles
    if show_y_axis == False:
        ax.axes.get_yaxis().set_visible(False)
        x_indent = 0.015
    else:
        x_indent = -0.005
    
    # Remove the y-axis label
    ax.set_ylabel(None)
    
    # Remove the x-axis label
    ax.set_xlabel(None)
    
    # Show the mean if requested
    if show_mean:
        # Calculate the mean
        mean = dataframe[value_column_name].mean()
        # Show the mean as a vertical line with a label
        ax.axvline(
            x=mean,
            ymax=0.97-.02,
            color="#262626",
            linestyle="--",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=mean, 
            y=plt.ylim()[1] * 0.97, 
            s='Mean: {:.2f}'.format(mean),
            horizontalalignment='center',
            # fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
    
    # Show the median if requested
    if show_median:
        # Calculate the median
        median = dataframe[value_column_name].median()
        # Show the median as a vertical line with a label
        ax.axvline(
            x=median,
            ymax=0.90-.02,
            color="#262626",
            linestyle=":",
            linewidth=1.5,
            alpha=0.5
        )
        ax.text(
            x=median,
            y=plt.ylim()[1] * .90,
            s='Median: {:.2f}'.format(median),
            horizontalalignment='center',
            # fontname="Arial",
            fontsize=9,
            color="#262626",
            alpha=0.75
        )
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
    # Set x-axis tick label font to Arial, size 9, and color #666666
    ax.tick_params(
        axis='x',
        which='major',
        labelsize=9,
        labelcolor="#666666",
        pad=2,
        bottom=True,
        labelbottom=True
    )
    # plt.xticks(fontname='Arial')
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
from math import ceil
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import random
import seaborn as sns
import textwrap

# Delcare function
def SimulateNormallyDistributedOutcome(expected_outcome=0,
                                       standard_deviation_of_outcome=None,
                                       min_max_of_outcome=None,
                                       # Simulation parameters
                                       number_of_trials=10000,
                                       return_format='dataframe',
                                       simulated_variable_name='Simulated Outcome',
                                       random_seed=412,
                                       # Plotting parameters
                                       plot_simulation_results=True,
                                       fill_color="#999999",
                                       fill_transparency=0.6,
                                       figure_size=(8, 6),
                                       show_mean=True,
                                       show_median=True,
                                       # Text formatting arguments
                                       title_for_plot="Simulation Results",
                                       subtitle_for_plot="Showing the distribution of the outcome",
                                       caption_for_plot=None,
                                       data_source_for_plot=None,
                                       show_y_axis=False,
                                       title_y_indent=1.1,
                                       subtitle_y_indent=1.05,
                                       caption_y_indent=-0.15):
    """
    Simulates normally distributed outcomes based on the specified parameters.
    The normal distribution is a continuous probability distribution that is symmetrical around its mean, most of the observations cluster around the central peak, and the probabilities for values further away from the mean taper off equally in both directions. 
    Extreme values in both tails of the distribution are similarly unlikely.
    Conditions:
    - Continuous data
    - Unbounded distribution
    - Outliers are minimal

    Args:
        expected_outcome (float): The expected value of the outcome. Defaults to 0.
        standard_deviation_of_outcome (float): The standard deviation of the outcome. If not specified, it is estimated using min_max_of_outcome. Defaults to None.
        min_max_of_outcome (list): A list of length 2 specifying the minimum and maximum values of the outcome. If specified, standard_deviation_of_outcome is ignored. Defaults to None.
        number_of_trials (int): The number of trials to simulate. Defaults to 10000.
        return_format (str): The format in which to return the simulation results. Must be either 'dataframe' or 'array'. Defaults to 'dataframe'.
        simulated_variable_name (str): The name of the simulated variable. Defaults to 'Simulated Outcome'.
        random_seed (int): The random seed to use for replicability. Defaults to 412.
        plot_simulation_results (bool): Whether to plot the simulation results. Defaults to True.
        fill_color (str): The color to use for the histogram fill. Defaults to "#999999".
        fill_transparency (float): The transparency of the histogram fill. Defaults to 0.6.
        figure_size (tuple): The size of the plot figure. Defaults to (8, 6).
        show_mean (bool): Whether to show the mean on the plot. Defaults to True.
        show_median (bool): Whether to show the median on the plot. Defaults to True.
        title_for_plot (str): The title of the plot. Defaults to "Simulation Results".
        subtitle_for_plot (str): The subtitle of the plot. Defaults to "Showing the distribution of the outcome".
        caption_for_plot (str): The caption of the plot. Defaults to None.
        data_source_for_plot (str): The data source of the plot. Defaults to None.
        show_y_axis (bool): Whether to show the y-axis on the plot. Defaults to False.
        title_y_indent (float): The y-indent of the plot title. Defaults to 1.1.
        subtitle_y_indent (float): The y-indent of the plot subtitle. Defaults to 1.05.
        caption_y_indent (float): The y-indent of the plot caption. Defaults to -0.15.

    Returns:
        pandas.DataFrame or numpy.ndarray: The simulated outcomes in the specified format.
    """
    
    # Ensure arguments are valid
    if standard_deviation_of_outcome is not None and standard_deviation_of_outcome < 0:
        raise ValueError("Please make sure that your standard_deviation_of_outcome argument is greater than or equal to 0.")
    
    # Ensure that return_format is either 'dataframe' or 'array'
    if return_format not in ['dataframe', 'array']:
        raise ValueError("return_format must be either 'dataframe' or 'array'.")
    
    # Ensure that min_max_of_outcome is either None or a list of length 2
    if min_max_of_outcome is not None:
        if len(min_max_of_outcome) != 2:
            raise ValueError("If specified, min_max_of_outcome must be a list of length 2.")
    
    # If specified, set random seed for replicability
    if random_seed is not None:
        random.seed(random_seed)
    
    # If standard_deviation_of_outcome and min_max_of_outcome are not None, print a warning
    if standard_deviation_of_outcome is not None and min_max_of_outcome is not None:
        print("Warning: Both standard_deviation_of_outcome and min_max_of_outcome were specified. Ignoring min_max_of_outcome.")
    
    # If standard_deviation_of_outcome is not specified, estimate it using min_max_of_outcome
    if standard_deviation_of_outcome is None:
        if min_max_of_outcome is None:
            raise ValueError("Please specify either standard_deviation_of_outcome or min_max_of_outcome.")
        else:
            # This formula is known as the "range rule" and it is based on the empirical finding that the range 
            # of a normal distribution is typically equal to 3.29 times its standard deviation.
            standard_deviation_of_outcome = (min_max_of_outcome[1] - min_max_of_outcome[0]) / 3.29  
        
    # Simulate normally distributed outcome
    list_sim_results = np.random.normal(
        loc=expected_outcome,
        scale=standard_deviation_of_outcome,
        size=number_of_trials
    )
    
    # Create dataframe of simulation results
    df_simulation = pd.DataFrame(list_sim_results,
                                 columns=[simulated_variable_name])
    
    # Generate plot if user requests it
    if plot_simulation_results == True:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        sns.histplot(
            data=df_simulation,
            x=simulated_variable_name,
            color=fill_color,
            alpha=fill_transparency,
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the mean if requested
        if show_mean:
            # Calculate the mean
            mean = df_simulation[simulated_variable_name].mean()
            # Show the mean as a vertical line with a label
            ax.axvline(
                x=mean,
                ymax=0.97-.02,
                color="#262626",
                linestyle="--",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=mean, 
                y=plt.ylim()[1] * 0.97, 
                s='Mean: {:.2f}'.format(mean),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Show the median if requested
        if show_median:
            # Calculate the median
            median = df_simulation[simulated_variable_name].median()
            # Show the median as a vertical line with a label
            ax.axvline(
                x=median,
                ymax=0.90-.02,
                color="#262626",
                linestyle=":",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=median,
                y=plt.ylim()[1] * .90,
                s='Median: {:.2f}'.format(median),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
        
    # Return the metalog distribution
    if return_format == 'dataframe':
        return df_simulation
    else:
        return np.array(list_sim_results)




# Load packages
import matplotlib.pyplot as plt
import numpy as np

# Declare function
def FindLimitOfFunction(f_of_x,
                        point,
                        step=0.0001,
                        plot_function=True,
                        plot_tangent_line=True,
                        x_minimum=-10,
                        x_maximum=10,
                        n=100,
                        tangent_line_window=None):
    """
    This function finds the x_tangent of a function at a given point. 
    It also plots the function and the tangent line at the point of interest.

    Args:
        f_of_x (lambda): The function of x.
        point (int or float): The point at which to find the x_tangent.
        step (float, optional): The step size to use when calculating the x_tangent. Defaults to 0.0001.
        plot_function (bool, optional): Whether or not to plot the function. Defaults to True.
        x_minimum (int, optional): The minimum value of x to use when plotting the function. Defaults to -10.
        x_maximum (int, optional): The maximum value of x to use when plotting the function. Defaults to 10.
        n (int, optional): The number of points to use when plotting the function. Defaults to 100.
        tangent_line_window (_type_, optional): The window to use when plotting the tangent line. Defaults to None, which will use 1/5 of the x range.
        
    Returns:
        float: The x_tangent at the point of interest.
    """
    
    # Create array of values based on the function
    x = np.linspace(x_minimum, x_maximum, n)
    y = np.zeros(len(x))
    for i in range(len(x)):
        y[i] = f_of_x(x[i])
    
    # Calculate the x_tangent
    x_tangent = (f_of_x(point+step) - f_of_x(point-step)) / (2*step)
    
    # Calculate the limit at the point of interest
    try:
        limit = f_of_x(point)
        print("The limit at x={0} is ~{1}".format(point, limit))
        # Create the tangent line
        y_tangent = f_of_x(point) + x_tangent*(x - point)
        
        # Plot the function if requested
        if plot_function:
            # Plot point at the point of interest
            plt.plot(point, f_of_x(point), "ro")
            
        # Plot the tangent line
        if plot_tangent_line:
            # plt.scatter(x0, my_function(x0), color='red')
            plt.plot(x, y_tangent, label='Tangent line', color='red', alpha=0.5)
            plt.scatter(point, f_of_x(point), color='red')
        
        # Plot the function if requested
        if plot_function:
            # Plot the function
            plt.plot(x, y, color="black", label='Function')
            
            # Add title
            plt.title("f(x)")
            
            # Add legend
            plt.legend()
            
            # Show plot
            plt.show()
            plt.clf()
        
        # Return the limit
        return limit
        
    except ZeroDivisionError:
        print("The limit at x={0} is undefined.".format(point))



# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def Plot100PercentStackedBarChart(dataframe,
                                  group_column_name='Group',
                                  value_column_name='Current Value',
                                  target_value_column_name=None,
                                  # Plot formatting arguments
                                  background_color='#e8e8ed',
                                  color_palette="Set1",
                                  fill_color=None,
                                  fill_transparency=0.8,
                                  figure_size=(8, 6),
                                  # Text formatting arguments
                                  data_label_fontsize=10,
                                  title_for_plot=None,
                                  subtitle_for_plot=None,
                                  caption_for_plot=None,
                                  data_source_for_plot=None,
                                  title_y_indent=1.15,
                                  subtitle_y_indent=1.1,
                                  caption_y_indent=-0.15):
    # Ensure that the number of unique values in the 'Group' column is equal to the number of rows in the DataFrame
    if len(dataframe[group_column_name].unique()) != len(dataframe):
        raise ValueError("The number of unique values in the 'Group' column must be equal to the number of rows in the DataFrame.")
    
    # Calculate the percentage of the Current Value out of the Target Value
    if target_value_column_name != None:
        dataframe['Percentage'] = dataframe[value_column_name] / dataframe[target_value_column_name] * 100
        dataframe['Remaining'] = 100 - dataframe['Percentage']
    else:
        dataframe['Percentage'] = dataframe[value_column_name] / dataframe[value_column_name].sum() * 100
        dataframe['Remaining'] = 100 - dataframe['Percentage']
    
    # Sort the DataFrame based on the 'Group' column to match the user's example
    dataframe.sort_values(by='Group', inplace=True)
    
    # Short by the 'Percentage' column to make the plot look better
    dataframe.sort_values(by='Percentage', inplace=True)
    
    # Set the figure size for better clarity
    plt.figure(figsize=figure_size)
    
    # If fill_color is not provided, use the color_palette
    if fill_color == None:
        # Create a color palette
        color_palette = sns.color_palette(color_palette, len(dataframe))
        
        # Create a list of colors
        colors = [color_palette[i] for i in range(len(dataframe))]
    else:
        # Create a list of colors
        colors = [fill_color for i in range(len(dataframe))]
        
    # Create subplot
    ax = plt.subplot(111)
    
    # Plot the 'Percentage' bar horizontally
    ax.barh(
        dataframe['Group'], 
        dataframe['Percentage'], 
        color=colors,
        edgecolor='white',
        alpha=fill_transparency
    )
    
    # Add the 'Remaining' bar horizontally behind the 'Percentage' bar
    ax.barh(
        dataframe['Group'], 
        dataframe['Remaining'], 
        left=dataframe['Percentage'], 
        color=background_color, 
        edgecolor='white',
    )
    
    # Set the x-axis to show percentages
    plt.xticks(np.arange(0, 101, 10))
    
    # Limit the x-axis to 100%
    plt.xlim(0, 100)
    
    # Add space between the title and the plot
    plt.subplots_adjust(top=0.85)
    
    # Wrap y axis label using textwrap
    wrapped_variable_name = "\n".join(textwrap.wrap(group_column_name, 40, break_long_words=False))  # String wrap the variable name
    plt.ylabel(wrapped_variable_name, fontsize=10, color="#262626")
    
    # Format and wrap y axis tick labels using textwrap
    wrapped_y_tick_labels = ['\n'.join(textwrap.wrap(label, 40, break_long_words=False)) for label in dataframe[group_column_name].tolist()]
    plt.yticks(
        dataframe['Group'], 
        wrapped_y_tick_labels, 
        fontsize=10, 
        color="#262626"
    )
    
    # Move x-axis to the top
    ax.xaxis.tick_top()
    
    # Change x-axis colors to "#666666"
    plt.tick_params(axis='x', colors="#666666")
    
    # Remove the top, right, and left spines
    ax.spines['bottom'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    
    # Set the x indent of the plot titles and captions
    # Get longest y tick label
    longest_y_tick_label = max(wrapped_y_tick_labels, key=len)
    if len(longest_y_tick_label) >= 30:
        x_indent = -0.3
    else:
        x_indent = -0.005 - (len(longest_y_tick_label) * 0.011)
        
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
      
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
        
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
from math import ceil, sqrt
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import random
import seaborn as sns
import textwrap

# Declare function
def SimulateTDistributedOutcome(degrees_of_freedom,
                                expected_outcome=0,
                                standard_deviation_of_outcome=None,
                                min_max_of_outcome=None,
                                number_of_trials=10000,
                                return_format='dataframe',
                                simulated_variable_name='Simulated Outcome',
                                random_seed=412,
                                plot_simulation_results=True,
                                fill_color="#999999",
                                fill_transparency=0.6,
                                figure_size=(8, 6),
                                show_mean=True,
                                show_median=True,
                                title_for_plot="Simulation Results",
                                subtitle_for_plot="Showing the distribution of the outcome",
                                caption_for_plot=None,
                                data_source_for_plot=None,
                                show_y_axis=False,
                                title_y_indent=1.1,
                                subtitle_y_indent=1.05,
                                caption_y_indent=-0.15):
    """
    Simulates a T-distributed outcome with the specified degrees of freedom and other parameters.
    The T distribution (also called Student's T Distribution) is a family of distributions that look almost identical to the normal distribution curve, only a bit shorter and fatter. 
    A Student's t-distribution is used where one would be inclined to use a Normal distribution, but a Normal distribution is susceptible to outliers whereas a t-distribution is more robust.
    Conditions:
    - Continuous data
    - Unbounded distribution
    - Considered an overdispersed Normal distribution, mixture of individual normal distributions with different variances

    Args:
        degrees_of_freedom (int): The degrees of freedom for the T distribution.
        expected_outcome (float, optional): The expected value of the outcome. Defaults to 0.
        standard_deviation_of_outcome (float, optional): The standard deviation of the outcome. If not specified, it is estimated using min_max_of_outcome. Defaults to None.
        min_max_of_outcome (list of 2 floats, optional): The minimum and maximum values of the outcome. If standard_deviation_of_outcome is not specified, it is estimated using this range. Defaults to None.
        number_of_trials (int, optional): The number of trials to simulate. Defaults to 10000.
        return_format (str, optional): The format in which to return the simulation results. Must be either 'dataframe' or 'array'. Defaults to 'dataframe'.
        simulated_variable_name (str, optional): The name of the simulated variable. Defaults to 'Simulated Outcome'.
        random_seed (int, optional): The random seed to use for replicability. Defaults to 412.
        plot_simulation_results (bool, optional): Whether to plot the simulation results. Defaults to True.
        fill_color (str, optional): The color to fill the histogram bars with. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the histogram bars. Defaults to 0.6.
        figure_size (tuple of 2 ints, optional): The size of the plot figure. Defaults to (8, 6).
        show_mean (bool, optional): Whether to show the mean on the plot. Defaults to True.
        show_median (bool, optional): Whether to show the median on the plot. Defaults to True.
        title_for_plot (str, optional): The title of the plot. Defaults to "Simulation Results".
        subtitle_for_plot (str, optional): The subtitle of the plot. Defaults to "Showing the distribution of the outcome".
        caption_for_plot (str, optional): The caption of the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis on the plot. Defaults to False.
        title_y_indent (float, optional): The y-indent of the title on the plot. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent of the subtitle on the plot. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent of the caption on the plot. Defaults to -0.15.

    Returns:
        pandas.DataFrame or numpy.ndarray: The simulated T-distributed outcome in the specified format.
    """
    
    # Ensure arguments are valid
    if standard_deviation_of_outcome is not None and standard_deviation_of_outcome < 0:
        raise ValueError("Please make sure that your standard_deviation_of_outcome argument is greater than or equal to 0.")
    if degrees_of_freedom <= 0:
        raise ValueError("Please make sure that your degrees_of_freedom argument is greater than or equal to 1.")
    
    # Ensure that return_format is either 'dataframe' or 'array'
    if return_format not in ['dataframe', 'array']:
        raise ValueError("return_format must be either 'dataframe' or 'array'.")
    
    # Ensure that min_max_of_outcome is either None or a list of length 2
    if min_max_of_outcome is not None:
        if len(min_max_of_outcome) != 2:
            raise ValueError("If specified, min_max_of_outcome must be a list of length 2.")
    
    # If specified, set random seed for replicability
    if random_seed is not None:
        random.seed(random_seed)
    
    # If standard_deviation_of_outcome and min_max_of_outcome are not None, print a warning
    if standard_deviation_of_outcome is not None and min_max_of_outcome is not None:
        print("Warning: Both standard_deviation_of_outcome and min_max_of_outcome were specified. Ignoring min_max_of_outcome.")
    
    # If standard_deviation_of_outcome is not specified, estimate it using min_max_of_outcome
    if standard_deviation_of_outcome is None:
        if min_max_of_outcome is None:
            raise ValueError("Please specify either standard_deviation_of_outcome or min_max_of_outcome.")
        else:
            standard_deviation_of_outcome = (min_max_of_outcome[1] - min_max_of_outcome[0]) / (4 * sqrt(10))
            # The range rule of a t-distribution is typically equal to 4 times the standard deviation of the distribution, 
            # multiplied by the square root of the degrees of freedom.
    
    # If specified, set random seed for replicability
    if random_seed is not None:
        random.seed(random_seed)
        
    # Simulate T distributed outcome
    list_sim_results = np.random.standard_t(df=degrees_of_freedom,
                                            size=number_of_trials)
    
    # Convert T score to original value scale
    list_sim_results = list_sim_results * standard_deviation_of_outcome + expected_outcome
    
    # Create dataframe of simulation results
    df_simulation = pd.DataFrame(list_sim_results,
                                 columns=[simulated_variable_name])
    
    # Generate plot if user requests it
    if plot_simulation_results == True:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        sns.histplot(
            data=df_simulation,
            x=simulated_variable_name,
            color=fill_color,
            alpha=fill_transparency,
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the mean if requested
        if show_mean:
            # Calculate the mean
            mean = df_simulation[simulated_variable_name].mean()
            # Show the mean as a vertical line with a label
            ax.axvline(
                x=mean,
                ymax=0.97-.02,
                color="#262626",
                linestyle="--",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=mean, 
                y=plt.ylim()[1] * 0.97, 
                s='Mean: {:.2f}'.format(mean),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Show the median if requested
        if show_median:
            # Calculate the median
            median = df_simulation[simulated_variable_name].median()
            # Show the median as a vertical line with a label
            ax.axvline(
                x=median,
                ymax=0.90-.02,
                color="#262626",
                linestyle=":",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=median,
                y=plt.ylim()[1] * .90,
                s='Median: {:.2f}'.format(median),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
        
    # Return the metalog distribution
    if return_format == 'dataframe':
        return df_simulation
    else:
        return np.array(list_sim_results)




# Import packages
from lida import llm, Manager, TextGenerationConfig
from lida.utils import plot_raster, plt
import openai
import pandas as pd
from matplotlib import pyplot as plt
import numpy as np
import seaborn as sns

# Declare function
def GenerateEDAWithLIDA(dataframe,
                        llm_api_key,
                        # LLM parameters
                        llm_provider="openai",
                        llm_model="gpt-3.5-turbo",
                        visualization_library="seaborn",
                        goal_temperature=0.50,
                        code_generation_temperature=0.05,
                        # Data summarization parameters
                        data_summary_method="llm",
                        number_of_samples_to_show_in_summary=5,
                        return_data_fields_summary=True,
                        # EDA parameters
                        number_of_goals_to_generate=5,
                        plot_recommended_visualization=False,
                        show_code_for_recommended_visualization=False):
    """
    This function generates exploratory data analysis (EDA) goals using the LIDA package from Microsoft.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe containing the data to be analyzed.
        llm_api_key (str): The API key for the LLM provider.
        llm_provider (str): The LLM provider. Defaults to "openai".
        llm_model (str): The LLM model. Defaults to "gpt-3.5-turbo".
        visualization_library (str): The visualization library to use. Defaults to "seaborn".
        goal_temperature (float, optional): The temperature to use for goal generation. Defaults to 0.50.
        code_generation_temperature (float, optional): The temperature to use for code generation. Defaults to 0.05.
        data_summary_method (str, optional): The method to use for data summarization. Defaults to "llm".
        number_of_samples_to_show_in_summary (int, optional): The number of samples to show in the data summary. Defaults to 5.
        return_data_fields_summary (bool, optional): Whether to return the data fields summary. Defaults to True.
        number_of_goals_to_generate (int, optional): The number of goals to generate. Defaults to 5.
        plot_recommended_visualization (bool, optional): Whether to plot the recommended visualization. Defaults to False.
        show_code_for_recommended_visualization (bool, optional): Whether to show the code for the recommended visualization. Defaults to False.

    Returns:
        Pandas dataframe: A dataframe containing the data fields summary.
    """
    # Set up LIDA and text generation model
    lida = Manager(text_gen=llm(
        provider=llm_provider,
        api_key=llm_api_key,
    )) 

    # Set up text generation config for goal generation
    goal_textgen_config = TextGenerationConfig(
        n=1, 
        temperature=goal_temperature, 
        model=llm_model, 
        use_cache=True
    )

    # Set up text generation config for code generation
    code_textgen_config = TextGenerationConfig(
        n=1,
        temperature=code_generation_temperature,
        model=llm_model,
        use_cache=True
    )

    # Create summary and goals
    summary = lida.summarize(
        data=dataframe, 
        summary_method=data_summary_method,
        textgen_config=goal_textgen_config,
        n_samples=number_of_samples_to_show_in_summary
    )
    goals = lida.goals(
        summary, 
        n=number_of_goals_to_generate,
        textgen_config=goal_textgen_config
    )
    
    # Print summary of the data
    if data_summary_method == "llm":
        print("-----  DATA SUMMARY  -----")
        # Get name of dataset from summary JSON
        print("Name:", summary['name'])
        print("Description:", summary['dataset_description'])
        
        # Convert fields to dataframe
        fields = pd.DataFrame(summary['fields'])
        
        # The properties column is a dictionary, so convert it to columns
        properties = pd.DataFrame(fields['properties'].tolist())
        fields = pd.concat([fields, properties], axis=1)
        fields = fields.drop(columns=['properties'])
        
        # Set column as index
        fields = fields.set_index('column')
        
        # Move description column to the front
        fields = fields[['description'] + [col for col in fields.columns if col != 'description']]
            

    # Create charts for goals
    print("-----  EXPLORATORY DATA ANALYSIS (EDA) GOALS  -----")
    for i in range(number_of_goals_to_generate):
        print("Question #" + str(i+1) + ":", goals[i].question)
        print("Why:", goals[i].rationale)
        if plot_recommended_visualization or show_code_for_recommended_visualization:
            # Generate charts
            charts = lida.visualize(
                summary=summary, 
                goal=goals[i], 
                textgen_config=goal_textgen_config, 
                library=visualization_library
            )
            
            # Show the plot if requested
            if plot_recommended_visualization:
                plot_raster(charts[0].raster)
                
            # Show the code if requested
            if show_code_for_recommended_visualization:
                print("Starter code for recommended visualization:", "\n\n", charts[0].code)
        else:
            # Describe the recommended visualization
            print("Recommended visulization:", goals[i].visualization)
        print("\n")
    
    # Return data summary dataframe if requested
    if return_data_fields_summary:
        return fields




# Load packages
from IPython.display import display, Markdown, HTML, Latex
import pandas as pd
from tableone import TableOne

# Declare function
def RenderTableOne(dataframe,
                   value_column_name,
                   grouping_column_name,
                   list_of_row_variables,
                   table_format='html',
                   show_p_value=True,
                   return_table_object=False):
    """
    Generates and displays a summary table (Table 1) from a given pandas DataFrame.

    Args:
        dataframe (pandas.DataFrame): The DataFrame that contains the data.
        value_column_name (str): The name of the column in the DataFrame to be used as the outcome variable.
        grouping_column_name (str): The name of the column in the DataFrame to be used as the grouping variable.
        list_of_row_variables (list): A list of column names to be included as rows in the table.
        table_format (str, optional): The format of the table. Defaults to 'html'.
        show_p_value (bool, optional): Whether to include p-values in the table. Defaults to True.
        return_table_object (bool, optional): Whether to return the TableOne object. Defaults to False.

    Returns:
        tableone.TableOne: The TableOne object, if return_table_object is True. Otherwise, None.
    """
    
    # Select the columns to be included in the table
    dataframe = dataframe[[value_column_name, grouping_column_name] + list_of_row_variables]
    
    # Create table 1 object
    table_one = TableOne(dataframe, 
                         columns=list_of_row_variables,
                         groupby=grouping_column_name, 
                         pval=show_p_value)
    
    # Render the table
    if table_format in ['html']:
        table_one_display = table_one.tabulate(
            tablefmt=table_format, 
            floatfmt=".2f"
        )
        table_one_display = HTML(table_one_display)
        display(table_one_display)
    elif table_format in ['grid', 'simple', 'github']:
        table_one_display = table_one.tabulate(
            tablefmt=table_format, 
            floatfmt=".2f"
        )
        table_one_display = Markdown(table_one_display)
        display(table_one_display)
    elif table_format in ['latex_booktabs', 'latex']:
        table_one_display = table_one.tabulate(
            tablefmt=table_format, 
            floatfmt=".2f"
        )
        table_one_display = Latex(table_one_display)
        display(table_one_display)
    else:
        table_one_display = table_one.tabulate(
            tablefmt=table_format, 
            floatfmt=".2f"
        )
        print(table_one_display)
    
    # Return the table object
    if return_table_object:
        return(table_one)




# Load packages
import os
import pandas as pd
import pymetalog as pm
from math import ceil
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def CreateMetalogDistributionFromPercentiles(list_of_values,
                                             list_of_percentiles,
                                             # Metalog distribution parameters
                                             lower_bound=None,
                                             upper_bound=None,
                                             learning_rate=.01,
                                             term_maximum=None,
                                             term_minimum=2,
                                             term_for_random_sample=None,
                                             number_of_samples=10000,
                                             variable_name="Simulated Value",
                                             show_summary=True,
                                             return_format='dataframe',
                                             # Plot parameters
                                             show_distribution_plot=True,
                                             figure_size=(8, 6),
                                             fill_color="#999999",
                                             fill_transparency=0.6,
                                             show_mean=True,
                                             show_median=True,
                                             show_y_axis=False,
                                             # Plot text parameters
                                             title_for_plot=None,
                                             subtitle_for_plot=None,
                                             caption_for_plot=None,
                                             data_source_for_plot=None,
                                             title_y_indent=1.1,
                                             subtitle_y_indent=1.05,
                                             caption_y_indent=-0.15):
    """
    Creates a metalog distribution from a list of values and a list of percentiles.

    Args:
        list_of_values (list): A list of values.
        list_of_percentiles (list): A list of percentiles.
        lower_bound (float, optional): The lower bound of the metalog distribution. Defaults to None.
        upper_bound (float, optional): The upper bound of the metalog distribution. Defaults to None.
        learning_rate (float, optional): The learning rate of the metalog distribution. Defaults to .01.
        term_maximum (int, optional): The maximum number of terms used in the metalog distribution. Defaults to None.
        term_minimum (int, optional): The minimum number of terms used in the metalog distribution. Defaults to 2.
        term_for_random_sample (int, optional): The number of terms used in the metalog distribution for a random sample. Defaults to None.
        number_of_samples (int, optional): The number of samples to take from the metalog distribution. Defaults to 10000.
        variable_name (str, optional): The name of the variable. Defaults to "Simulated Value".
        show_summary (bool, optional): Whether to show the summary of the metalog distribution. Defaults to True.
        return_format (str, optional): The format of the return value. Must be one of "dataframe" or "array". Defaults to "dataframe".
        show_distribution_plot (bool, optional): Whether to show the metalog distribution plot. Defaults to True.
        figure_size (tuple, optional): The size of the metalog distribution plot. Defaults to (8, 6).
        fill_color (str, optional): The fill color of the metalog distribution plot. Defaults to "#999999".
        fill_transparency (float, optional): The fill transparency of the metalog distribution plot. Defaults to 0.6.
        show_mean (bool, optional): Whether to show the mean on the metalog distribution plot. Defaults to True.
        show_median (bool, optional): Whether to show the median on the metalog distribution plot. Defaults to True.
        show_y_axis (bool, optional): Whether to show the y-axis on the metalog distribution plot. Defaults to False.
        title_for_plot (str, optional): The title of the metalog distribution plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle of the metalog distribution plot. Defaults to None.
        caption_for_plot (str, optional): The caption of the metalog distribution plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the metalog distribution plot. Defaults to None.
        title_y_indent (float, optional): The y-indent of the title on the metalog distribution plot. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent of the subtitle on the metalog distribution plot. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent of the caption on the metalog distribution plot. Defaults to -0.15.

    Returns:
        pandas.DataFrame or numpy.ndarray: The metalog distribution in the specified format.
    """
    
    # Ensure that the list of values and list of percentiles are the same length
    if len(list_of_values) != len(list_of_percentiles):
        raise ValueError("The list of values and list of percentiles must be the same length.")
    
    # Ensure that return_format is a valid argument
    if return_format not in ["dataframe", "array"]:
        raise ValueError("return_format must be one of the following: dataframe or array.")
    
    # If term_maximum is not provided, set it to the length of the list of values
    if term_maximum is None:
        term_maximum = len(list_of_values)
        
    # Create a metalog distribution
    if lower_bound is None and upper_bound is None:
        metalog_dist = pm.metalog(
            x=list_of_values,
            probs=list_of_percentiles,
            step_len=learning_rate,
            term_lower_bound=term_minimum,
            term_limit=term_maximum
        )
    elif lower_bound is not None and upper_bound is None:
        metalog_dist = pm.metalog(
            x=list_of_values,
            probs=list_of_percentiles,
            boundedness='sl',
            bounds=[lower_bound],
            step_len=learning_rate,
            term_lower_bound=term_minimum,
            term_limit=term_maximum
        )
    elif lower_bound is None and upper_bound is not None:
        metalog_dist = pm.metalog(
            x=list_of_values,
            probs=list_of_percentiles,
            boundedness='su',
            bounds=[upper_bound],
            step_len=learning_rate,
            term_lower_bound=term_minimum,
            term_limit=term_maximum
        )
    else:
        metalog_dist = pm.metalog(
            x=list_of_values,
            probs=list_of_percentiles,
            boundedness='b',
            bounds=[lower_bound, upper_bound],
            step_len=learning_rate,
            term_lower_bound=term_minimum,
            term_limit=term_maximum
        )

    # Show summary of the metalog distribution
    if show_summary:
        pm.summary(m = metalog_dist)
        
    # Get the maximum number of terms used in the metalog distribution that is valid
    if term_for_random_sample is None:
        term_for_random_sample = metalog_dist.term_limit

    # Take a random sample from the metalog distribution
    arr_metalog = pm.rmetalog(
        m=metalog_dist, 
        n=number_of_samples,
        term=term_for_random_sample
    )
    
    # Convert the array to a dataframe
    metalog_df = pd.DataFrame(arr_metalog, columns=[variable_name])
    
    # Plot the metalog distribution
    if show_distribution_plot:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        sns.histplot(
            data=metalog_df,
            x=variable_name,
            color=fill_color,
            alpha=fill_transparency,
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the mean if requested
        if show_mean:
            # Calculate the mean
            mean = metalog_df[variable_name].mean()
            # Show the mean as a vertical line with a label
            ax.axvline(
                x=mean,
                ymax=0.97-.02,
                color="#262626",
                linestyle="--",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=mean, 
                y=plt.ylim()[1] * 0.97, 
                s='Mean: {:.2f}'.format(mean),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Show the median if requested
        if show_median:
            # Calculate the median
            median = metalog_df[variable_name].median()
            # Show the median as a vertical line with a label
            ax.axvline(
                x=median,
                ymax=0.90-.02,
                color="#262626",
                linestyle=":",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=median,
                y=plt.ylim()[1] * .90,
                s='Median: {:.2f}'.format(median),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
                
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
    
    # Return the metalog distribution
    if return_format == 'dataframe':
        return metalog_df
    else:
        return arr_metalog
    



# Load packages
import pandas as pd

# Declare function
def AddRowCountColumn(dataframe,
                      list_of_grouping_variables,
                      list_of_order_columns,
                      list_of_ascending_order_args=None,
                      row_count_column_name='Row Count'):
    """
    This function adds a row count column to a dataset based on a list of order columns, a list of grouping variables, and a list of ascending order arguments.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe
        list_of_grouping_variables (list): The list of columns to group by.
        list_of_order_columns (list): The list of columns to order by.
        list_of_ascending_order_args (list, optional): The list of ascending order arguments. Defaults to None. If None, all columns are ordered in ascending order.
        row_count_column_name (str, optional): The name of the now row count column. Defaults to 'Row Count'.

    Returns:
        Pandas dataframe: An updated Pandas dataframe with a row count column.
    """
    
    # Order dataframe by order columns
    if (list_of_ascending_order_args == None):
        dataframe = dataframe.sort_values(
            list_of_order_columns
        )
    else:
        dataframe = dataframe.sort_values(
            list_of_order_columns,
            ascending=list_of_ascending_order_args
        )
        
    # Add row count
    dataframe = dataframe.groupby(
        list_of_grouping_variables
    ).cumcount()
    
    # Resort by grouping columns and row order column
    list_of_order_columns.append(row_count_column_name)
    dataframe = dataframe.sort_values(
        list_of_order_columns
    )
    
    # Return dataframe
    return(dataframe)




# Load packages
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler
import seaborn as sns
from sklearn import metrics
from sklearn.model_selection import train_test_split
import tensorflow as tf

# Declare function
def CreateContentBasedRecommender(dataframe,
                                  outcome_variable,
                                  user_list_of_predictor_variables,
                                  item_list_of_predictor_variables,
                                  user_number_of_hidden_layers=2,
                                  item_number_of_hidden_layers=2,
                                  number_of_recommendations=10,
                                  test_size=0.2,
                                  scale_variables=True,
                                  plot_loss=True,
                                  plot_model_test_performance=True,
                                  show_predictor_ranges=False,
                                  initial_learning_rate=0.01,
                                  number_of_steps_gradient_descent=50,
                                  lambda_for_regularization=0.001,
                                  random_seed=412):
    
    # Keep only the predictors from the user and item dataframes
    dataframe = dataframe[user_list_of_predictor_variables + item_list_of_predictor_variables + [outcome_variable]].copy()
    
    # Set the random seed
    tf.random.set_seed(random_seed)
    
    # Use Keras to create a normalization layer
    if scale_variables:
        # For users
        scalerUser = StandardScaler()
        scalerUser.fit(dataframe[user_list_of_predictor_variables])
        dataframe[user_list_of_predictor_variables] = scalerUser.transform(dataframe[user_list_of_predictor_variables])
        # For items
        scalerItem = StandardScaler()
        scalerItem.fit(dataframe[item_list_of_predictor_variables])
        dataframe[item_list_of_predictor_variables] = scalerItem.transform(dataframe[item_list_of_predictor_variables])
        # Outcome variable
        scalerOutcome = MinMaxScaler((-1, 1))
        scalerOutcome.fit(dataframe[outcome_variable].values.reshape(-1,1))
        dataframe[outcome_variable] = scalerOutcome.transform(dataframe[outcome_variable].values.reshape(-1,1))
        
    # Show the peak-to-peak range of each predictor
    if show_predictor_ranges:
        print("\nPeak-to-peak range of each predictor in user dataset:")
        print(np.ptp(dataframe[user_list_of_predictor_variables], axis=0))
        print("\nPeak-to-peak range of each predictor in item dataset:")
        print(np.ptp(dataframe[item_list_of_predictor_variables], axis=0))
    
    # Split dataframe into training and test sets
    train, test = train_test_split(
        dataframe, 
        test_size=test_size,
        shuffle=True,
        random_state=random_seed
    )
    
    # Create dictionary of layers to be used in the neural network
    user_dict_layers = {}
    item_dict_layers = {}
    
    # Create hidden layers
    # # For users
    for i in range(user_number_of_hidden_layers):
        key_text = 'Hidden layer ' + str(i + 1)
        if i < 10:
            layer_name = 'layer_0' + str(i + 1)
        else:
            layer_name = 'layer_' + str(i + 1)
        user_dict_layers[key_text] = tf.keras.layers.Dense(
            number_of_recommendations + ((user_number_of_hidden_layers - 1 - i) * 10),
            activation='relu', 
            name=layer_name,
            kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
        )
    # # For items
    for i in range(item_number_of_hidden_layers):
        key_text = 'Hidden layer ' + str(i + 1)
        if i < 10:
            layer_name = 'layer_0' + str(i + 1)
        else:
            layer_name = 'layer_' + str(i + 1)
        item_dict_layers[key_text] = tf.keras.layers.Dense(
            number_of_recommendations + ((item_number_of_hidden_layers - 1 - i) * 10),
            activation='relu', 
            name=layer_name,
            kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
        )
        
    # Create output layer
    # # For users
    user_dict_layers['Output layer'] = tf.keras.layers.Dense(
        number_of_recommendations, 
        activation='linear', 
        name='final_layer',
        kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
    )
    # # For items
    item_dict_layers['Output layer'] = tf.keras.layers.Dense(
        number_of_recommendations, 
        activation='linear', 
        name='final_layer',
        kernel_regularizer=tf.keras.regularizers.l2(lambda_for_regularization)
    )
    
    # Create list of layers to be used in the neural networks
    # # For users
    user_list_of_layers = []
    for i in range(user_number_of_hidden_layers):
        key_text = 'Hidden layer ' + str(i + 1)
        user_list_of_layers.append(user_dict_layers[key_text])
    user_list_of_layers.append(user_dict_layers['Output layer'])
    # # For items
    item_list_of_layers = []
    for i in range(item_number_of_hidden_layers):
        key_text = 'Hidden layer ' + str(i + 1)
        item_list_of_layers.append(item_dict_layers[key_text])
    item_list_of_layers.append(item_dict_layers['Output layer'])
    
    # Create sequential models
    user_model = tf.keras.Sequential(user_list_of_layers)
    item_model = tf.keras.Sequential(item_list_of_layers)
    
    # Create input layers and point to base layers
    # # For users
    user_input = tf.keras.layers.Input(shape=(len(user_list_of_predictor_variables),))
    user_output = user_model(user_input)
    user_output = tf.linalg.l2_normalize(user_output, axis=1)
    # # For items
    item_input = tf.keras.layers.Input(shape=(len(item_list_of_predictor_variables),))
    item_output = item_model(item_input)
    item_output = tf.linalg.l2_normalize(item_output, axis=1)
    
    # Create recommendation output layer
    recommender_output = tf.keras.layers.Dot(axes=1)([user_output, item_output])
    
    # Create model
    model = tf.keras.Model([user_input, item_input], recommender_output)
    
    # Show model summary
    model.summary()
    
    # Define loss function and optimizer
    model.compile(
        loss=tf.keras.losses.MeanSquaredError(),
        optimizer=tf.keras.optimizers.Adam(learning_rate=initial_learning_rate),
        metrics=['mean_squared_error']
    )
    
    # Train the model
    loss_history = model.fit(
        [train[user_list_of_predictor_variables].values, train[item_list_of_predictor_variables].values],
        train[outcome_variable].values,
        epochs=number_of_steps_gradient_descent,
        verbose=0
    )
    if plot_loss:
        plt.figure(figsize=(9,9))
        sns.lineplot(
            x=range(1, number_of_steps_gradient_descent + 1),
            y=loss_history.history['loss']
        )
        plt.title('Loss Curve', size = 15)
        plt.xlabel('Epoch')
        plt.ylabel('Loss')
        plt.show()
        
    # # Evaluate the model
    # model.evaluate(
    #     [test[user_list_of_predictor_variables].values, test[item_list_of_predictor_variables].values],
    #     test[outcome_variable].values
    # )
    
    # Make predictions on test set
    predictions = model.predict([test[user_list_of_predictor_variables].values, test[item_list_of_predictor_variables].values])
    if scale_variables:
        predictions = scalerOutcome.inverse_transform(predictions.reshape(-1, 1))
        test[outcome_variable] = scalerOutcome.inverse_transform(test[outcome_variable].values.reshape(-1, 1))
    test['Predicted'] = predictions
    
    # Plot prediction performance if requested
    if plot_model_test_performance:
        plt.figure(figsize=(9,9))
        if len(dataframe[outcome_variable].unique()) == 2:
            sns.scatterplot(
                data=test,
                x=outcome_variable,
                y='Predicted'
            )
            plt.title('Probability vs. Observed Outcome', size = 15)
            plt.ylabel('Predicted probability')
            plt.ylim(0, 1.05)
        else:
            sns.regplot(
                data=test,
                x=outcome_variable,
                y='Predicted'
            )
            plt.plot(test[outcome_variable], test[outcome_variable], color='black', alpha=0.35)
            plt.title('Predicted vs. Observed Outcome', size = 15)
        plt.show()
        
    # Create dictionary of objects to be returned
    if scale_variables:
        dict_return = {
            'User Scaler': scalerUser,
            'Item Scaler': scalerItem,
            'Outcome Scaler': scalerOutcome,
            'Model': model
        }
        return(dict_return)
    else:
        return(model)




# Load packages
from math import ceil
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import random
import seaborn as sns
import textwrap

# Declare function
def SimulateCountUntilFirstSuccess(probability_of_success,
                                   # Simulation parameters
                                   number_of_trials=10000,
                                   simulated_variable_name='Count Until First Success',
                                   random_seed=412,
                                   return_format='dataframe',
                                   # Plotting parameters
                                   plot_simulation_results=True,
                                   fill_color="#999999",
                                   fill_transparency=0.6,
                                   figure_size=(8, 6),
                                   show_mean=True,
                                   show_median=True,
                                   # Text formatting arguments
                                   title_for_plot="Simulation Results",
                                   subtitle_for_plot="Showing the distribution of the count until first success",
                                   caption_for_plot=None,
                                   data_source_for_plot=None,
                                   show_y_axis=False,
                                   title_y_indent=1.1,
                                   subtitle_y_indent=1.05,
                                   caption_y_indent=-0.15):
    """
    Simulate the count until the first success using a negative binomial distribution.
    A negative binomial distribution can be used to describe the number of successes r - 1
    and x failures in x + r -1 trials, until you have a success on the x + rth trial. 
    Rephrased, this models the number of failures (x) you would have to see before you see a 
    certain number of successes (r).
    Conditions:
    - Count of discrete events
    - The events CAN be non-independent, implying that events can influence or cause other events
    - Variance can exceed the mean

    Args:
        probability_of_success (float): The probability of success for each trial.
        number_of_trials (int, optional): The number of trials to simulate. Defaults to 10000.
        simulated_variable_name (str, optional): The name of the simulated variable. Defaults to 'Count Until First Success'.
        random_seed (int, optional): The random seed for replicability. Defaults to 412.
        return_format (str, optional): The format of the output. Either 'dataframe' or 'array'. Defaults to 'dataframe'.
        plot_simulation_results (bool, optional): Whether to plot the simulation results. Defaults to True.
        fill_color (str, optional): The color to fill the histogram bars with. Defaults to "#999999".
        fill_transparency (float, optional): The transparency of the histogram bars. Defaults to 0.6.
        figure_size (tuple, optional): The size of the plot figure. Defaults to (8, 6).
        show_mean (bool, optional): Whether to show the mean on the plot. Defaults to True.
        show_median (bool, optional): Whether to show the median on the plot. Defaults to True.
        title_for_plot (str, optional): The title of the plot. Defaults to "Simulation Results".
        subtitle_for_plot (str, optional): The subtitle of the plot. Defaults to "Showing the distribution of the count until first success".
        caption_for_plot (str, optional): The caption of the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source of the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis on the plot. Defaults to False.
        title_y_indent (float, optional): The y-indent of the plot title. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent of the plot subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent of the plot caption. Defaults to -0.15.

    Returns:
        pandas.DataFrame or numpy.ndarray: The simulated count until the first success.
    """
    
    # Ensure probability_of_success is between 0 and 1
    if probability_of_success > 1 or probability_of_success < 0:
        raise ValueError("Please change your probability_of_success argument -- it must be greater than 0 and less than 1.")
    
    # Ensure that return_format is either 'dataframe' or 'array'
    if return_format not in ['dataframe', 'array']:
        raise ValueError("return_format must be either 'dataframe' or 'array'.")
    
    # If specified, set random seed for replicability
    if random_seed is not None:
        random.seed(random_seed)

    # Simulate count until first success
    list_sim_results = []
    for i in range(0, number_of_trials):
        is_success = False
        event_count = 0
        while is_success == False:
            event_count += 1
            sim_result = random.choices(population = [0, 1],
                                        weights = [1-probability_of_success, probability_of_success])
            sim_result = sum(sim_result)
            if sim_result == 1:
                list_sim_results.append(event_count)
                is_success = True
            else:
                continue
    df_simulation = pd.DataFrame(list_sim_results,
                                 columns = [simulated_variable_name])
    
    # Generate plot if user requests it
    if plot_simulation_results == True:
        # Create figure and axes
        fig, ax = plt.subplots(figsize=figure_size)
        
        # Create histogram using seaborn
        sns.histplot(
            data=df_simulation,
            x=simulated_variable_name,
            color=fill_color,
            alpha=fill_transparency,
        )
        
        # Remove top, left, and right spines. Set bottom spine to dark gray.
        ax.spines['top'].set_visible(False)
        ax.spines['left'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color("#262626")
        
        # Remove the y-axis, and adjust the indent of the plot titles
        if show_y_axis == False:
            ax.axes.get_yaxis().set_visible(False)
            x_indent = 0.015
        else:
            x_indent = -0.005
        
        # Remove the y-axis label
        ax.set_ylabel(None)
        
        # Remove the x-axis label
        ax.set_xlabel(None)
        
        # Show the mean if requested
        if show_mean:
            # Calculate the mean
            mean = df_simulation[simulated_variable_name].mean()
            # Show the mean as a vertical line with a label
            ax.axvline(
                x=mean,
                ymax=0.97-.02,
                color="#262626",
                linestyle="--",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=mean, 
                y=plt.ylim()[1] * 0.97, 
                s='Mean: {:.2f}'.format(mean),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Show the median if requested
        if show_median:
            # Calculate the median
            median = df_simulation[simulated_variable_name].median()
            # Show the median as a vertical line with a label
            ax.axvline(
                x=median,
                ymax=0.90-.02,
                color="#262626",
                linestyle=":",
                linewidth=1.5,
                alpha=0.5
            )
            ax.text(
                x=median,
                y=plt.ylim()[1] * .90,
                s='Median: {:.2f}'.format(median),
                horizontalalignment='center',
                fontname="Arial",
                fontsize=9,
                color="#262626",
                alpha=0.75
            )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Set x-axis tick label font to Arial, size 9, and color #666666
        ax.tick_params(
            axis='x',
            which='major',
            labelsize=9,
            labelcolor="#666666",
            pad=2,
            bottom=True,
            labelbottom=True
        )
        plt.xticks(fontname='Arial')
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
        # Show plot
        plt.show()
        
        # Clear plot
        plt.clf()
        
    # Return the metalog distribution
    if return_format == 'dataframe':
        return df_simulation
    else:
        return np.array(list_sim_results)




# Load packages
from PyPDF2 import PdfReader, PdfWriter

# Declare function
def CreateCopyOfPDF(input_file, 
                    output_file, 
                    start_page=None, 
                    end_page=None):
    """
    This function creates a copy of a PDF file.

    Args:
        input_file (str): The path to the input PDF file to copy (including the file name).
        output_file (str): The path to the output PDF file to create (including the file name).
        start_page (int, optional): The page number to start copying from. Defaults to None (first page).
        end_page (int, optional): The page number to stop copying at. Defaults to None (last page).
    """
    # If start_page is None, set to 0
    if start_page is None:
        start_page = 1
    
    # If end_page is None, set to last page of PDF
    if end_page is None:
        end_page = PdfFileReader(input_file).getNumPages()
    
    # If start_page is greater than end_page, raise error
    if start_page > end_page:
        raise ValueError('start_page must be less than or equal to end_page')
    
    # Read in the input PDF
    with open(input_file, 'rb') as input_pdf:
        pdf_reader = PdfReader(input_pdf)
        pdf_writer = PdfWriter()
        
        # Add the correct pages to the output PDF
        for page in range(start_page, end_page + 1):
            pdf_writer.add_page(pdf_reader.pages[page - 1])
        
        # Write the output PDF to a file
        with open(output_file, 'wb') as output_pdf:
            pdf_writer.write(output_pdf)




# Load packages
import pandas as pd
from matplotlib import pyplot as plt
import seaborn as sns
import textwrap

# Declare function
def PlotBoxWhiskerByGroup(dataframe,
                          value_column_name,
                          grouping_column_name_1,
                          grouping_column_name_2=None,
                          # Plot formatting arguments
                          fill_color=None,
                          color_palette='Set2',
                          display_order_list=None,
                          show_legend=True,
                          # Text formatting arguments
                          title_for_plot=None,
                          subtitle_for_plot=None,
                          caption_for_plot=None,
                          data_source_for_plot=None,
                          show_y_axis=False,
                          title_y_indent=1.1,
                          subtitle_y_indent=1.05,
                          caption_y_indent=-0.15,
                          x_indent=-0.10,
                          figure_size=(8, 6)):
    """
    Function to create a box and whisker plot for a given outcome variable, grouped by one or two categorical variables. 

    Args:
        dataframe (pandas.DataFrame): The input dataframe.
        value_column_name (str): The name of the outcome variable.
        grouping_column_name_1 (str): The name of the first group variable.
        grouping_column_name_2 (str, optional): The name of the second group variable. Defaults to None.
        fill_color (str, optional): The color to fill the box plot with. Defaults to None.
        color_palette (str, optional): The color palette to use for the box plot. Defaults to 'Set2'.
        display_order_list (list, optional): The order to display the 1st categories in the plot. Defaults to None.
        title_for_plot (str, optional): The title for the plot. Defaults to None.
        subtitle_for_plot (str, optional): The subtitle for the plot. Defaults to None.
        caption_for_plot (str, optional): The caption for the plot. Defaults to None.
        data_source_for_plot (str, optional): The data source for the plot. Defaults to None.
        show_y_axis (bool, optional): Whether to show the y-axis. Defaults to False.
        title_y_indent (float, optional): The y-indent for the title. Defaults to 1.1.
        subtitle_y_indent (float, optional): The y-indent for the subtitle. Defaults to 1.05.
        caption_y_indent (float, optional): The y-indent for the caption. Defaults to -0.15.
        x_indent (float, optional): The x-indent for the plot. Defaults to -0.128.
        figure_size (tuple, optional): The size of the plot. Defaults to (8, 6).
    """
    # If display_order_list is provided, check that it contains all of the categories in the dataframe
    if display_order_list != None:
        if not set(display_order_list).issubset(set(dataframe[grouping_column_name_1].unique())):
            raise ValueError("display_order_list must contain all of the categories in the " + grouping_column_name_1 + " column of dataframe.")
    else:
        # If display_order_list is not provided, create one from the dataframe
        display_order_list = dataframe.sort_values(value_column_name, ascending=True)[grouping_column_name_1].unique()
    
    # Create figure and axes
    fig, ax = plt.subplots(figsize=figure_size)
    
    # Generate box whisker plot
    if grouping_column_name_2 != None:
        if fill_color != None:
            ax = sns.boxplot(
                data=dataframe,
                x=grouping_column_name_1,
                y=value_column_name, 
                hue=grouping_column_name_2, 
                color=fill_color,
                order=display_order_list,
            )
        else:
            ax = sns.boxplot(
                data=dataframe,
                x=grouping_column_name_1,
                y=value_column_name, 
                hue=grouping_column_name_2, 
                palette=color_palette,
                order=display_order_list,
            )
    else:
        if fill_color != None:
            ax = sns.boxplot(
                data=dataframe,
                x=grouping_column_name_1,
                y=value_column_name, 
                color=fill_color,
                order=display_order_list,
            )
        else:
            ax = sns.boxplot(
                data=dataframe,
                x=grouping_column_name_1,
                y=value_column_name, 
                palette=color_palette,
                order=display_order_list,
            )
    
    # Remove top, and right spines. Set bottom and left spine to dark gray.
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_color("#262626")
    ax.spines['left'].set_color("#262626")
    
    # If show_legend is False, remove the legend
    if show_legend == False:
        ax.legend_.remove()
    
    # Add space between the title and the plot
    plt.subplots_adjust(top=0.85)
    
    # Set the title with Arial font, size 14, and color #262626 at the top of the plot
    ax.text(
        x=x_indent,
        y=title_y_indent,
        s=title_for_plot,
        # fontname="Arial",
        fontsize=14,
        color="#262626",
        transform=ax.transAxes
    )
    
    # Set the subtitle with Arial font, size 11, and color #666666
    ax.text(
        x=x_indent,
        y=subtitle_y_indent,
        s=subtitle_for_plot,
        # fontname="Arial",
        fontsize=11,
        color="#666666",
        transform=ax.transAxes
    )
    
    # String wrap group variable 1 tick labels
    group_variable_1_tick_labels = ax.get_xticklabels()
    group_variable_1_tick_labels = [label.get_text() for label in group_variable_1_tick_labels]
    for label in group_variable_1_tick_labels:
        label = textwrap.fill(label, 30, break_long_words=False)
    ax.set_xticklabels(group_variable_1_tick_labels)
    
    # Set x-axis tick label font to Arial, size 9, and color #666666
    ax.tick_params(
        axis='x',
        which='major',
        labelsize=9,
        labelcolor="#666666",
        pad=2,
        bottom=True,
        labelbottom=True
    )
    # plt.xticks(fontname='Arial')
    
    # Set y-axis tick label font to Arial, size 9, and color #666666
    ax.tick_params(
        axis='y',
        which='major',
        labelsize=9,
        labelcolor="#666666",
        pad=2,
        bottom=True,
        labelbottom=True
    )
    # plt.yticks(fontname='Arial')
    
    # Add a word-wrapped caption if one is provided
    if caption_for_plot != None or data_source_for_plot != None:
        # Create starting point for caption
        wrapped_caption = ""
        
        # Add the caption to the plot, if one is provided
        if caption_for_plot != None:
            # Word wrap the caption without splitting words
            wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
            
        # Add the data source to the caption, if one is provided
        if data_source_for_plot != None:
            wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
        
        # Add the caption to the plot
        ax.text(
            x=x_indent,
            y=caption_y_indent,
            s=wrapped_caption,
            # fontname="Arial",
            fontsize=8,
            color="#666666",
            transform=ax.transAxes
        )
    
    # Show plot
    plt.show()
    
    # Clear plot
    plt.clf()




# Load packages
from matplotlib import pyplot as plt
from matplotlib.patches import Rectangle
from mlxtend.frequent_patterns import apriori, association_rules
from mlxtend.preprocessing import TransactionEncoder
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap

# Declare function
def CreateAssociationRules(dataframe,
                           transaction_id_column,
                           items_column,
                           support_threshold=.01,
                           confidence_threshold=.05,
                           # Plotting parameters
                           plot_lift=True,
                           dot_fill_color="#999999",
                           upper_left_quadrant_fill_color="#32a852",
                           lower_right_quadrant_fill_color="#d14a41",
                           draw_life_equals_1_line=True,
                           # Plot text formatting arguments
                           title_for_plot="Association Rules",
                           subtitle_for_plot="Confidence vs. Consequent Support",
                           caption_for_plot="Lift is a measure of how much more likely it is for two items to occur together than expected by chance. A lift value of 1 means that the two items are independent of each other, while a lift value greater than 1 means that the two items are positively correlated, and a lift value less than 1 means that the two items are negatively correlated.",
                           data_source_for_plot=None,
                           x_indent=-0.128,
                           title_y_indent=1.125,
                           subtitle_y_indent=1.05,
                           caption_y_indent=-0.3,):
    """
    This function creates association rules from a dataset.
    
    Args:
        dataframe (Pandas dataframe): Pandas dataframe containing the data to be analyzed.
        transaction_id_column (str): The column name containing the transaction ID.
        items_column (str): The column name containing the items.
        support_threshold (float, optional): The threshold for the support for an association rule to be defined. Defaults to .20.
        confidence_threshold (float, optional): The threshold for the confidence for an association rule to be defined. Defaults to .50.
    
    Returns:
        Pandas dataframe: An updated Pandas dataframe with the association rules.
    """
    # Ensure that the support threshold is between 0 and 1
    if support_threshold < 0 or support_threshold > 1:
        raise ValueError("Support threshold must be between 0 and 1.")
    
    # Ensure that the confidence threshold is between 0 and 1
    if confidence_threshold < 0 or confidence_threshold > 1:
        raise ValueError("Confidence threshold must be between 0 and 1.")
    
    # Group and summarize (as list) items to key column
    dataframe = dataframe.groupby(transaction_id_column)[items_column].apply(list)
    dataframe = dataframe.reset_index()

    # Create association rule mining-ready dataset of transactions
    transactions = TransactionEncoder()
    df_transactions = transactions.fit(dataframe[items_column]).transform(dataframe[items_column])
    df_transactions = pd.DataFrame(df_transactions,
                                columns=transactions.columns_)

    # Create apriori model
    if support_threshold != None:
        df_association_rules = apriori(
            df_transactions, 
            min_support = support_threshold,
            use_colnames = True, 
            verbose = 1
        )
    else:
        df_association_rules = apriori(
            df_transactions,
            use_colnames = True, 
            verbose = 1
        )

    # Create association rules based on apriori model
    df_association_rules = association_rules(
        df_association_rules, 
        metric = "confidence", 
        min_threshold = confidence_threshold
    )
    df_association_rules = df_association_rules.sort_values(
        by=['confidence', 'consequents'],
        ascending=[False, True]
    )
    
    # Create scatter plot of association rules with support on the x-axis and confidence on the y-axis
    if plot_lift:
        ax = sns.scatterplot(
            data=df_association_rules, 
            x="consequent support",
            y="confidence",
            color=dot_fill_color,
            alpha=0.5,
            linewidth=0.5,
            edgecolor=dot_fill_color
        )
        
        # Draw a line where support and confidence are equal
        if draw_life_equals_1_line:
            ax.plot(
                [0, 1], 
                [0, 1],
                linestyle='--',
                color='#666666',
                linewidth=0.5
            )
            
        # Shade area where lift is greater than 1
        ax.fill_between(
            [0, 1], 
            [0, 1], 
            1,
            facecolor=upper_left_quadrant_fill_color,
            alpha=0.15
        )
        
        # Show area where lift is less than 1
        ax.fill_between(
            [0, 1], 
            [0, 1], 
            0,
            facecolor=lower_right_quadrant_fill_color,
            alpha=0.15
        )
        
        # Remove top and right spines, and set bottom and left spines to gray
        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.spines['bottom'].set_color('#666666')
        ax.spines['left'].set_color('#666666')
        
        # Format tick labels to be Arial, size 9, and color #666666
        ax.tick_params(
            which='major',
            labelsize=9,
            color='#666666'
        )
        
        # Find the middle of the x-axis and y-axis
        x_start = ax.get_xlim()[0]
        x_end = ax.get_xlim()[1]
        y_start = ax.get_ylim()[0]
        y_end = ax.get_ylim()[1]
        x_middle = x_start + (x_end - x_start) / 2
        y_middle = y_start + (y_end - y_start) / 2
        
        # Get coordinates for upper left quadrant label
        upper_left_quadrant_label_x = x_start + (x_middle - x_start) / 2
        upper_left_quadrant_label_x = upper_left_quadrant_label_x + (x_start - upper_left_quadrant_label_x) / 2
        upper_left_quadrant_label_y = y_middle + (y_end - y_middle) / 2
        upper_left_quadrant_label_y = upper_left_quadrant_label_y + (y_end - upper_left_quadrant_label_y) / 2

        # Write text in upper left quadrant
        ax.text(
            upper_left_quadrant_label_x, 
            upper_left_quadrant_label_y, 
            "Higher Lift", 
            fontsize=10,
            fontweight='bold',
            color=upper_left_quadrant_fill_color,
            alpha=0.8,
            ha='center', 
            va='center'
        )
        
        # Get coordinates for lower right quadrant label
        lower_right_quadrant_label_x = x_middle + (x_end - x_middle) / 2
        lower_right_quadrant_label_x = lower_right_quadrant_label_x + (x_end - lower_right_quadrant_label_x) / 2
        lower_right_quadrant_label_y = y_start + (y_middle - y_start) / 2
        lower_right_quadrant_label_y = lower_right_quadrant_label_y - (y_middle - lower_right_quadrant_label_y) / 2
    
        # Write text in lower right quadrant
        ax.text(
            lower_right_quadrant_label_x, 
            lower_right_quadrant_label_y, 
            "Lower Lift", 
            fontsize=10,
            fontweight='bold',
            color=lower_right_quadrant_fill_color,
            alpha=0.8,
            ha='center', 
            va='center'
        )
        
        # Set the title with Arial font, size 14, and color #262626 at the top of the plot
        ax.text(
            x=x_indent,
            y=title_y_indent,
            s=title_for_plot,
            fontname="Arial",
            fontsize=14,
            color="#262626",
            transform=ax.transAxes
        )
        
        # Set the subtitle with Arial font, size 11, and color #666666
        ax.text(
            x=x_indent,
            y=subtitle_y_indent,
            s=subtitle_for_plot,
            fontname="Arial",
            fontsize=11,
            color="#666666",
            transform=ax.transAxes
        )
        
        # Move the y-axis label to the top of the y-axis, and set the font to Arial, size 9, and color #666666
        ax.yaxis.set_label_coords(-0.1, 0.84)
        ax.yaxis.set_label_text(
            "Confidence",
            fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Move the x-axis label to the right of the x-axis, and set the font to Arial, size 9, and color #666666
        ax.xaxis.set_label_coords(0.85, -0.1)
        ax.xaxis.set_label_text(
            "Consequent Support",
            fontname="Arial",
            fontsize=10,
            color="#666666"
        )
        
        # Add a word-wrapped caption if one is provided
        if caption_for_plot != None or data_source_for_plot != None:
            # Create starting point for caption
            wrapped_caption = ""
            
            # Add the caption to the plot, if one is provided
            if caption_for_plot != None:
                # Word wrap the caption without splitting words
                wrapped_caption = textwrap.fill(caption_for_plot, 110, break_long_words=False)
                
            # Add the data source to the caption, if one is provided
            if data_source_for_plot != None:
                wrapped_caption = wrapped_caption + "\n\nSource: " + data_source_for_plot
            
            # Add the caption to the plot
            ax.text(
                x=x_indent,
                y=caption_y_indent,
                s=wrapped_caption,
                fontname="Arial",
                fontsize=8,
                color="#666666",
                transform=ax.transAxes
            )
            
    # Reformat column names to title case
    df_association_rules.columns = df_association_rules.columns.str.title()

    # Return association rules
    return(df_association_rules)

